import numpy as np
from .. import const


def surface_potential_from_polar_log_g(polar_log_g, mass):
    """
    calculation of surface potential based from polar gravity acceleration

    :param polar_log_g: float; polar gravity acceleration
    :param mass: float; stellar mass
    :return:
    """
    polar_gravity_acceleration = np.power(10, polar_log_g)
    return - np.power(const.G * mass * polar_gravity_acceleration, 0.5)


def potential_fn(radius, *args):
    """
    implicit potential function

    :param radius: (np.)float; spherical variable
    :param args: (array), float; (polar angles) and desired value of potential
    :return:
    """
    return potential(radius, *args[0]) - args[1]


def potential(radius, *args):
    """
    function calculates potential on the given point of the star

    :param radius: (np.)float; spherical variable
    :param args: Tuple; (A, B) - such that: Psi = -a/r - b*r**2
    :return: (np.)float
    """
    a, b = args

    return - a / radius - b * np.power(radius, 2.0)


def pre_calculate_for_potential_value(*args, return_as_tuple=False):
    """
    Function calculates auxiliary values for calculation of primary component potential,
    and therefore they don't need to be wastefully recalculated every iteration in solver.

    :param return_as_tuple: return coefficients as a tuple of numpy vectors instead of numpy matrix
    :type return_as_tuple: bool
    :param args: Tuple; (mass, angular velocity of rotation, latitude angle (0, pi))
    :return: Tuple; (a, b) such that: Psi = -a/r - b*r**2
    """
    mass, angular_velocity, theta, = args

    a = const.G * mass
    b = 0.5 * np.power(angular_velocity * np.sin(theta), 2)

    if np.isscalar(theta):
        return a, b
    else:
        aa = a * np.ones(np.shape(theta))
        return (aa, b) if return_as_tuple else np.column_stack((aa, b))


def radial_potential_derivative(radius, *args):
    """
    function calculate radial derivative of potential function in spherical coordinates

    :param radius: float; radius of given point(s) in spherical coordinates
    :param args: Tuple; (A, B) - such that: Psi = -a/r**2 - 2*b*r
    :type args: Tuple;
    :return:
    """
    a, b = args

    return a / np.power(radius, 2) - 2 * b * radius




import numpy as np

from . transform import PulsationModeProperties
from .. import utils, const as c, units
from .. logger import getLogger
from .. import settings

logger = getLogger('pulse.mode')


class PulsationMode(object):
    """
    Pulsation mode data container.
    """
    MANDATORY_KWARGS = ["l", "m", "amplitude", "frequency"]

    OPTIONAL_KWARGS = ["start_phase", 'mode_axis_theta', 'mode_axis_phi', 'temperature_perturbation_phase_shift',
                       'horizontal_to_radial_amplitude_ratio', 'temperature_amplitude_factor', 'tidally_locked']
    ALL_KWARGS = MANDATORY_KWARGS + OPTIONAL_KWARGS

    def __init__(self, **kwargs):
        utils.invalid_kwarg_checker(kwargs=kwargs, kwarglist=PulsationMode.ALL_KWARGS, instance=self.__class__)
        utils.check_missing_kwargs(PulsationMode.MANDATORY_KWARGS, kwargs, instance_of=PulsationMode)
        kwargs = self.transform_input(**kwargs)

        # get logger
        logger.info(f"initialising object {self.__class__.__name__}")
        logger.debug(f"setting property components of class instance {self.__class__.__name__}")

        # self._n = np.nan
        self.l = np.nan
        self.m = np.nan
        self.amplitude = np.nan
        self.frequency = np.nan
        self.start_phase = 0
        self.mode_axis_theta = 0
        self.mode_axis_phi = 0
        self.temperature_perturbation_phase_shift = settings.DEFAULT_TEMPERATURE_PERTURBATION_PHASE_SHIFT
        self.horizontal_to_radial_amplitude_ratio = None
        self.tidally_locked = False
        self.horizontal_to_radial_amplitude_ratio = None
        self.temperature_amplitude_factor = None
        # phase shift in radians between surface geometry
        # perturbation and temperature perturbations

        # dimensionless normalized amplitudes
        self.radial_amplitude = None  # in distance units
        self.horizontal_amplitude = None  # in distance units

        # surface related aux variables
        self.points = None  # rotated spherical coordinates aligned with pulsation axis

        self.point_harmonics = None
        self.point_harmonics_derivatives = None
        self.complex_displacement = None
        self.tilt_phi = None
        self.tilt_theta = None

        self.init_properties(**kwargs)

        self.angular_frequency = c.FULL_ARC * self.frequency

        # spherical harmonics renormalization constant to rms = 1
        # TODO: this is a constant
        self.renorm_const = 2 * c.PI ** 0.5
        self.validate_mode()

    @property
    def default_input_units(self):
        """
        Returns set of default units of intialization parameters, in case, when provided without units.

        :return: elisa.units.DefaultPulsationsInputUnits;
        """
        return units.DefaultPulsationsInputUnits

    @property
    def default_internal_units(self):
        """
        Returns set of internal units of Star parameters.

        :return: elisa.units.DefaultPulsationsUnits;
        """
        return units.DefaultPulsationsUnits

    @staticmethod
    def transform_input(**kwargs):
        """
        Transform and validate input kwargs.

        :param kwargs: Dict;
        :return: Dict;
        """
        return PulsationModeProperties.transform_input(**kwargs)

    def validate_mode(self):
        if np.abs(self.m) > self.l:
            raise ValueError(f'Absolute value of azimuthal order m: {self.m} cannot '
                             f'be higher than degree of the mode l: {self.l}.')

    def init_properties(self, **kwargs):
        """
        Setup system properties from input.

        :param kwargs: Dict; all supplied input properties
        """
        logger.debug(f"initialising properties of PulsationMode, values: {kwargs}")
        for kwarg in kwargs:
            setattr(self, kwarg, kwargs[kwarg])



import json

from elisa.analytics.binary_fit.shared import eval_constraint_in_dict
from elisa.analytics.params import parameters


class FitResultHandler(object):
    """
    Handling of fit results in standard JSON format.
    """
    def __init__(self):
        self.result = None
        self.flat_result = None

    def get_result(self):
        """
        Returns model parameters in standard dict (JSON) format.

        :return: Dict; model parameters in a standardized format
        """
        return self.result

    def load_result(self, path, autofill_sma=False):
        """
        Function loads a JSON file containing model parameters and stores it as an attribute of AnalyticsTask fitting
        instance. This is useful if you want to examine already calculated results using functionality provided by the
        AnalyticsTask instances (e.g: LCBinaryAnalyticsTask, RVBinaryAnalyticsTask, etc.).

        :param path: str; location of a JSON file with parameters
        :param autofill_sma; bool; if True, the semi-major axis will be autofilled to fitting
                                   parameters if absent
        """
        with open(path, 'r') as f:
            loaded_result = json.load(f)
        self.set_result(loaded_result, autofill_sma=autofill_sma)

    def save_result(self, path):
        """
        Save result as JSON file.

        :param path: str; path to file
        """
        if self.result is None:
            raise IOError("No result to store.")

        with open(path, 'w') as f:
            json.dump(self.result, f, separators=(',', ': '), indent=4)

    def set_result(self, result, autofill_sma=False):
        """
        Set model parameters in dictionary (JSON format) as an attribute of AnalyticsTask fitting instance. This is
        useful f you want to examine already calculated results using functionality provided by the AnalyticsTask
        instances (e.g: LCBinaryAnalyticsTask, RVBinaryAnalyticsTask, etc.).

        :param result: Dict; model parameters in JSON format
        :param autofill_sma; bool; if True, function will try to autofill the semi-major axis to fitting
                                   parameters if absent
        """
        result = eval_constraint_in_dict(result)
        result = parameters.extend_result_with_sma(result) if autofill_sma else result
        self.result = result
        self.flat_result = parameters.deserialize_result(self.result)



import numpy as np

from .. import umpy as up, settings
from .. logger import getLogger

logger = getLogger('opt.fsolver')


def fsolve(func, x0, args=(), fprime=None, full_output=False,
           col_deriv=False, xtol=1.49012e-8, maxfev=0, band=None,
           epsfcn=None, factor=100, diag=None):
    solution, a, ier, b = up.optimize.fsolve(func, x0, args=args, fprime=fprime, full_output=True,
                                             col_deriv=col_deriv, xtol=xtol, maxfev=maxfev, band=band,
                                             epsfcn=epsfcn, factor=factor, diag=diag)
    if not full_output:
        return solution
    return solution, a, ier, b


def fsolver(fn, condition, *args, **kwargs):
    """
    Will solve `fn` implicit function taking args by using scipy.optimize.fsolve method and return
    solution if satisfy conditional function.

    :param fn: function;
    :param condition: function;
    :param args: Tuple;
    :param kwargs: Dict;
        * **original_kwargs** *
    :return: Tuple;
    """
    # precalculation of auxiliary values
    solution, use = np.nan, False
    scipy_solver_init_value = np.array([1e-4])
    try:
        solution, _, ier, msg = fsolve(fn, scipy_solver_init_value, full_output=True, args=args, xtol=1e-10)
        if ier == 1 and not up.isnan(solution[0]):
            solution = solution[0]
            use = True if 1e15 > solution > 0 else False
        else:
            if not settings.SUPPRESS_WARNINGS:
                logger.warning(f'solution in implicit solver was not found, cause: {msg}')
    except Exception as e:
        logger.debug(f"attempt to solve function {fn.__name__} finished w/ exception: {str(e)}")
        # noinspection PyUnusedLocal
        use = False
        raise

    args_to_use = kwargs.get('original_kwargs', args)
    return (solution, use) if condition(solution, *args_to_use) else (np.nan, False)



from ... import utils
from ... base.surface.coverage import surface_area_coverage
from ... logger import getLogger

logger = getLogger('single_system.surface.coverage')


def compute_surface_coverage(system):
    """
    Compute surface coverage of faces for given rotational position
    defined by container/SystemContainer.

    :param system:
    :return:
    """
    logger.debug(f"computing surface coverage for {system.position}")
    star = getattr(system, 'star')

    coverage = utils.poly_areas(star.points[star.faces[star.indices]])
    coverage = surface_area_coverage(len(star.faces), star.indices, coverage)

    return {'star': coverage, }



import hashlib


def md5_content(content):
    md5 = hashlib.md5()
    content = content.encode('utf-8') if isinstance(content, str) else content
    md5.update(content)
    return md5.hexdigest()


def sha256_content(content):
    sha256 = hashlib.sha256()
    content = content.encode('utf-8') if isinstance(content, str) else content
    sha256.update(content)
    return sha256.hexdigest()


def det_2d(matrix):
    return matrix[0][0] * matrix[1][1] - matrix[0][1] * matrix[1][0]


def multiple_determinants(matrix):
    """
    Calculates 2D determinant on every level of given 3D matrix.
    :param matrix: np.array (Nx2x2), where i-th slice looks like::

        [[xi1, yi1], [xi2, yi2]]

    :return: numpy.array; N-dim vector where each element is 2D determinant of
                          2 2D vectors stored on given level in `matrix`
    """
    return matrix[:, 0, 0] * matrix[:, 1, 1] - matrix[:, 0, 1] * matrix[:, 1, 0]



import numpy as np


def ballesteros_formula(color_bv):
    """
    Ballesteros formula.
    Estimate start temperature from B-V index.

    color_bv: B - V = -2.5 * log(F_b / F_v)

    :param color_bv: float; B-V index
    :return: float; temperature stimation
    """
    return 4600.0 * ((1.0 / ((0.92 * color_bv) + 1.7)) + (1.0 / ((0.92 * color_bv) + 0.62)))


def pogsons_formula(f1, f2):
    """
    return -2.5 * log(f1 / f2)

    :param f1: Union[float, numpy.array];
    :param f2: Union[float, numpy.array];
    :return: Union[float, numpy.array];
    """

    return -2.5 * np.log10(f1 / f2)


def _overcontact_temperature_estimation(b_v):
    a, b, c = 1270.92384801, 1.73588834, 1290.17377233
    return (a / np.log10((b + b_v))) + c


def _solar_like_temperature_estimation(b_v):
    a, b, c = 1768.60111726, 1.78240258, -264.71844987
    return (a / np.log10((b + b_v))) + c


def elisa_bv_temperature(b_v, morphology="detached"):
    """
    Temperature estimation for elisa based on B-V index.

    :param b_v: Union[float, numpy.array]; B-V index/indices
    :param morphology: str;
    :return: Union[float, numpy.array]; temperature/s estimation
    """

    if morphology in ['overcontact']:
        return _overcontact_temperature_estimation(b_v)
    return _solar_like_temperature_estimation(b_v)



import numpy as np

from collections.abc import Sequence
from matplotlib import pyplot as plt

from ... import umpy as up
from ... logger import getLogger
from ... utils import is_empty, polar_to_cartesian

logger = getLogger("orbit.container")


class OrbitalSupplements(Sequence):
    """
    Structure designed to store orbital positions on (nearly) symmetrical orbital positions around apsidal line.
    The symmetrical counterparts are stored in the same positions within 'body' and 'mirror' position arrays.
    Corresponding `body` and `mirror` items share the same bindary system model which saves computational time.
    !!! BEWARE, THIS IS MUTABLE !!!

    :param body: numpy.array; N x 5 array containing row-wise orbital positions used as templates
                             (they will be evaluated exactly)
    :param mirror: numpy.array; N x 5 array containing row-wise orbital positions mirrored from corresponding
                               `body` orbital positions using symmetry along apsidal line.
    """

    def __getitem__(self, index):
        return self.body[index], self.mirror[index]

    def __init__(self, body=None, mirror=None):
        if body is None and mirror is None:
            self._body = np.array([])
            self._mirror = np.array([])

        else:
            self._body = np.array(body)
            self._mirror = np.array(mirror)

    def append(self, body, mirror):
        """
        Appending corresponding `body`/`mirror` couple to the end of their respective arrays.

        :param body: numpy.array; [index, ] + [items for items in elisa.const.Position]
        :param mirror: numpy.array; [index, ] + [items for items in elisa.const.Position]
        """
        self._body = np.vstack((self._body, body)) if not is_empty(self._body) else np.array([body])
        self._mirror = np.vstack((self._mirror, mirror)) if not is_empty(self._mirror) else np.array([mirror])

    @property
    def body(self):
        """Returns an array of body positions."""
        return self._body

    @property
    def mirror(self):
        """Returns an array of mirror positions."""
        return self._mirror

    @property
    def body_defined(self):
        """Returns True if `body` position array is not empty."""
        return self.not_empty(self.body)

    @property
    def mirror_defined(self):
        """Returns True if `mirror` position array is not empty."""
        return self.not_empty(self.mirror)

    @staticmethod
    def is_empty(val):
        """Return True if given orbital position is empty."""
        return np.all(up.isnan(val))

    @classmethod
    def not_empty(cls, arr):
        """
        Return values where supplied array is not empty.

        :param arr: numpy.array;
        :return: numpy.array;
        """
        return arr[list(map(lambda x: not cls.is_empty(x), arr))]

    def sort(self, by='distance'):
        """
        Sort by given quantity.
        This method sorts bodies and mirrors based on quantity chosen on input.
        Sorting of mirrors is based on sorting of bodies.

        :param by: str; `index`, `distance` or `radius` sorting orbital positions in `body` an `mirror` according to
                         the index of `body` positions or by component distances.
        :return: self; sorted OrbitalSupplements object
        """

        if by == 'index':
            by = 0
        elif by == 'distance' or by == 'radius':
            by = 1
        else:
            raise ValueError("Invalid value of `by`")

        sort_index = np.argsort(self.body[:, by])
        self._body = self.body[sort_index]
        self._mirror = self.mirror[sort_index]

        return self

    def size(self):
        return self.__len__()

    def to_orbital_position(self):
        pass

    def plot_bodies(self):
        """
        Visualize orbital positions stored in `body` .
        """
        self._plot(self.body_defined)

    def plot_mirrors(self):
        """
        Visualize orbital positions stored in `mirror` .
        """
        self._plot(self.mirror_defined, markers=["x", "x"])

    def plot(self):
        """
        Visualize orbital positions stored in OrbitalSupplements.
        """
        self._plot(self.body_defined, self.mirror_defined)

    @classmethod
    def _plot(cls, arr1, arr2=None, markers=None):
        markers = ["o", "x"] if markers is None else markers
        x, y = polar_to_cartesian(arr1[:, 1], arr1[:, 2] - (up.pi / 2))
        plt.scatter(x, y, marker=markers[0])

        if not is_empty(arr2):
            x, y = polar_to_cartesian(arr2[:, 1], arr2[:, 2] - (up.pi / 2))
            plt.scatter(x, y, marker=markers[1])

        plt.grid(True)
        plt.axes().set_aspect('equal')
        plt.show()

    def __iter__(self):
        for body, mirror in zip(self.body, self.mirror):
            yield body, mirror

    def __len__(self):
        return len(self.body)

    def __eq__(self, other):
        return np.all(self._body == other.body) & \
               np.all((self.mirror == other.mirror)[~np.all(up.isnan(other.mirror) & up.isnan(self.mirror), axis=1)])

    def __str__(self):
        return f"{self.__class__.__name__}\nbodies: {self.body}\nmirrors: {self._mirror}"

    def __repr__(self):
        return self.__str__()



import datetime
import re

import numpy as np
import pandas as pd
import scipy as sp

from typing import Sized
from queue import Empty
from numpy.linalg import norm
from matplotlib.cbook import flatten

from copy import copy, deepcopy
from . import const, umpy as up

from . numba_functions import operations
from . base.types import FLOAT, INT


def polar_to_cartesian(radius, phi):
    """
    Transform polar coordinates to cartesian.

    :param radius: (numpy.)float, (numpy.)int;
    :param phi: (numpy.)float, (numpy.)int;
    :return: Tuple ((numpy.)float, (numpy.)float);
    """
    x = radius * up.cos(phi)
    y = radius * up.sin(phi)
    return x, y


def invalid_kwarg_checker(kwargs, kwarglist, instance):
    """
    Check if `kwargs` for `instance` are in allowed kwargs presented in `kwarglist`.

    :param kwargs: Dict; kwargs to evaluate if are in kwarg list
    :param kwarglist: Dict;
    :param instance: Any; class/instance with attribute `__name__`
    """
    invalid_kwargs = [kwarg for kwarg in kwargs if kwarg not in kwarglist]
    if len(invalid_kwargs) > 0:
        raise ValueError(f'Invalid keyword argument(s): {", ".join(invalid_kwargs)} '
                         f'in class instance {instance.__name__}.\n '
                         f'List of available parameters: {", ".join(kwarglist)}')


def invalid_param_checker(kwargs, kwarglist, message):
    """

    :param kwargs: Dict; kwargs to evaluate if are in kwarg list
    :param kwarglist: Dict;
    :param message: Any class
    """
    invalid_kwargs = [kwarg for kwarg in kwargs if kwarg not in kwarglist]
    if len(invalid_kwargs) > 0:
        raise ValueError(f'Invalid keyword argument(s): {", ".join(invalid_kwargs)} '
                         f'in {message}.\n '
                         f'List of available parameters: {", ".join(kwarglist)}')


def is_plane(given, expected):
    """
    Find out whether `given` plane definition of 2d plane is `expected` one. E.g. if `yx` is `yx` or `xy`

    :param given: str;
    :param expected: str;
    :return: bool;
    """
    pattern = r'^({0})|({1})$'.format(expected, expected[::-1])
    return re.search(pattern, given)


def find_nearest_dist_3d(data):
    """
    Function finds the smallest distance between given set of 3D points.

    :param data: Iterable;
    :return: float; minimal distance of points in dataset
    """
    if isinstance(data, np.ndarray):
        data = data.tolist()
    from scipy.spatial import KDTree
    points = copy(data)
    test_points, distances = copy(points), []

    for i in range(0, len(test_points) - 1):
        points.remove(test_points[i])
        tree = KDTree(points)

        if np.isnan(test_points[i]).any():
            distances.append(np.inf)
        else:
            distance, ndx = tree.query([test_points[i]], k=1)
            distances.append(distance[0])
    return np.min(distances)


def cartesian_to_spherical(points, degrees=False):
    """
    Convert cartesian to spherical coordinates.
    If only 1 point is given input an output is only 1D vector

    :param points:

        shape::

                    numpy.array([[x1, y1, z1],
                                [x2, y2, z2],
                                 ...
                                [xn, yn, zn]])

    :param degrees: bool; whether return spherical angular coordinates in degrees
    :return: numpy.array;

    shhape::

        ([[r1, phi1, theta1],
          [r2, phi2, theta2],
           ...
          [rn, phin, thetan]])
    """
    points = np.array(points)
    points = np.expand_dims(points, axis=0) if len(np.shape(points)) == 1 else points
    r = np.linalg.norm(points, axis=1)

    old_settings = np.seterr(divide='ignore', invalid='ignore')
    phi = up.arcsin(points[:, 1] / (np.linalg.norm(points[:, :2], axis=1)))  # vypocet azimutalneho (rovinneho) uhla
    phi[up.isnan(phi)] = 0

    theta = up.arccos(points[:, 2] / r)  # vypocet polarneho (elevacneho) uhla
    theta[up.isnan(theta)] = 0
    np.seterr(**old_settings)

    signtest = points[:, 0] < 0
    phi[signtest] = (const.PI - phi[signtest])

    return_val = np.column_stack((r, phi, theta)) if not degrees else np.column_stack((r, up.degrees(phi),
                                                                                       up.degrees(theta)))
    return np.squeeze(return_val, axis=0) if np.shape(return_val)[0] == 1 else return_val


def cartesian_to_polar(points, degrees=False):
    points = np.insert(points, 2, np.zeros(len(points)), axis=1)
    transform = cartesian_to_spherical(points, degrees=degrees)
    return transform.T[:2].T


def spherical_to_cartesian(spherical_points):
    """
    Converts spherical coordinates into cartesian.
    If input is one point, output is 1D vector.
    Function assumes that supplied spherical angular coordinates are in radians.

    :param spherical_points: numpy.array;

    shape::

        numpy.array([[r1, phi1, theta1],
                     [r2, phi2, theta2],
                      ...
                     [rn, phin, thetan]])

    :return: numpy.array;

    shape::

        ([[x1, y1, z1],
          [x2, y2, z2],
          ...
          [xn, yn, zn]])
    """
    spherical_points = np.array(spherical_points)
    spherical_points = np.expand_dims(spherical_points, axis=0) if len(np.shape(spherical_points)) == 1 \
        else spherical_points
    x = spherical_points[:, 0] * up.cos(spherical_points[:, 1]) * up.sin(spherical_points[:, 2])
    y = spherical_points[:, 0] * up.sin(spherical_points[:, 1]) * up.sin(spherical_points[:, 2])
    z = spherical_points[:, 0] * up.cos(spherical_points[:, 2])
    points = np.column_stack((x, y, z))
    return np.squeeze(points, axis=0) if np.shape(points)[0] == 1 else points


def cylindrical_to_cartesian(cylindrical_points):
    """
    Converts cylindrical coordinates into cartesian.
    If input is one point, output is 1D vector.
    Function assumes that supplied spherical angular coordinates are in radians.

    :param cylindrical_points: numpy.array;

    shape ::

        numpy.array([[r1, phi1, z1],
                     [r2, phi2, z2],
                      ...
                     [rn, phin, zn]])

    :return: numpy.array;

    shape::

        ([[x1, y1, z1],
          [x2, y2, z2],
          ...
          [xn, yn, zn]])
    """
    cylindrical_points = np.array(cylindrical_points)
    cylindrical_points = np.expand_dims(cylindrical_points, axis=0) if len(np.shape(cylindrical_points)) == 1 \
        else cylindrical_points
    x = cylindrical_points[:, 0] * up.cos(cylindrical_points[:, 1])
    y = cylindrical_points[:, 0] * up.sin(cylindrical_points[:, 1])
    points = np.column_stack((x, y, cylindrical_points[:, 2]))
    return np.squeeze(points, axis=0) if np.shape(points)[0] == 1 else points


def arbitrary_rotation(theta, omega, vector, degrees=False, omega_normalized=False):
    """
    Rodrigues's Rotation Formula.
    Function rotates `vector` around axis defined by `omega` vector by amount `theta`.

    :param theta: float; radial vector of point of interest to rotate
    :param omega: Union[List, numpy.array]; 3d list of floats; arbitrary vector to rotate around
    :param vector: Union[List, numpy.array]; 3d list of floats;
    :param degrees: bool; if True, theta supplied on intput is in degrees otherwise in radians
    :param omega_normalized: bool; if True, then in-function normalization of omega is not performed
    :return: numpy.array;
    """
    # this action normalizes the same vector over and over again during spot calculation, which is unnecessary
    if not omega_normalized:
        omega = np.array(omega) / np.linalg.norm(np.array(omega))

    theta = theta if not degrees else up.radians(theta)

    matrix = up.arange(9, dtype=FLOAT).reshape((3, 3))

    matrix[0, 0] = (up.cos(theta)) + (omega[0] ** 2 * (1. - up.cos(theta)))
    matrix[1, 0] = (omega[0] * omega[1] * (1. - up.cos(theta))) - (omega[2] * up.sin(theta))
    matrix[2, 0] = (omega[1] * up.sin(theta)) + (omega[0] * omega[2] * (1. - up.cos(theta)))

    matrix[0, 1] = (omega[2] * up.sin(theta)) + (omega[0] * omega[1] * (1. - up.cos(theta)))
    matrix[1, 1] = (up.cos(theta)) + (omega[1] ** 2 * (1. - up.cos(theta)))
    matrix[2, 1] = (- omega[0] * up.sin(theta)) + (omega[1] * omega[2] * (1. - up.cos(theta)))

    matrix[0, 2] = (- omega[1] * up.sin(theta)) + (omega[0] * omega[2] * (1. - up.cos(theta)))
    matrix[1, 2] = (omega[0] * up.sin(theta)) + (omega[1] * omega[2] * (1. - up.cos(theta)))
    matrix[2, 2] = (up.cos(theta)) + (omega[2] ** 2 * (1. - up.cos(theta)))

    return up.matmul(vector, matrix)


def around_axis_rotation(theta, vector, axis, inverse=False, degrees=False):
    """
    Rotation of `vector` around `axis` by an amount `theta`.

    :param theta: float; degree of rotation
    :param vector: numpy.array; vector to rotate around
    :param axis: str; axis of rotation `x`, `y`, or `z`
    :param inverse: bool; rotate to inverse direction than is math positive
    :param degrees: bool; if True value theta is assumed to be in degrees
    :return: numpy.array; rotated vector(s)
    """
    matrix = up.arange(9, dtype=FLOAT).reshape((3, 3))
    theta = theta if not degrees else up.radians(theta)
    vector = np.array(vector)

    if axis == "x":
        matrix[0][0], matrix[1][0], matrix[2][0] = 1, 0, 0
        matrix[0][1], matrix[1][1], matrix[2][1] = 0, up.cos(theta), - up.sin(theta)
        matrix[0][2], matrix[1][2], matrix[2][2] = 0, up.sin(theta), up.cos(theta)
        if inverse:
            matrix[2][1], matrix[1][2] = up.sin(theta), - up.sin(theta)
    if axis == "y":
        matrix[0][0], matrix[1][0], matrix[2][0] = up.cos(theta), 0, up.sin(theta)
        matrix[0][1], matrix[1][1], matrix[2][1] = 0, 1, 0
        matrix[0][2], matrix[1][2], matrix[2][2] = - up.sin(theta), 0, up.cos(theta)
        if inverse:
            matrix[0][2], matrix[2][0] = + up.sin(theta), - up.sin(theta)
    if axis == "z":
        matrix[0][0], matrix[1][0], matrix[2][0] = up.cos(theta), - up.sin(theta), 0
        matrix[0][1], matrix[1][1], matrix[2][1] = up.sin(theta), up.cos(theta), 0
        matrix[0][2], matrix[1][2], matrix[2][2] = 0, 0, 1
        if inverse:
            matrix[1][0], matrix[0][1] = + up.sin(theta), - up.sin(theta)
    return up.matmul(vector, matrix)


def rotate_item(vector, position, inclination):
    """
    Transfer vector(s) from corotating reference frame to observers frame.

    :param vector: numpy.array;
    :param position: elisa.const.Position;
    :param inclination: float;
    :return: numpy.array
    """
    correction = np.sign(const.LINE_OF_SIGHT[0]) * const.HALF_PI
    args = (position.azimuth - correction, vector, "z", False, False)
    vector = around_axis_rotation(*args)

    inverse = False if const.LINE_OF_SIGHT[0] == 1 else True
    args = (const.HALF_PI - inclination, vector, "y", inverse, False)
    return around_axis_rotation(*args)


def average_spacing_cgal(data, neighbours=6):
    """
    Average Spacing - calculates average distance between points using average distances to `neighbours` number of
    points.

    Match w/ CGAL average spacing function (https://www.cgal.org/).

    :param data: List; 3-dimensinal dataset
    :param neighbours: int; nearest neighbours to average
    :return: float;
    """
    if not isinstance(data, type(np.array)):
        data = np.array(data)

    # noinspection PyUnresolvedReferences
    dist = sp.spatial.distance.cdist(data, data, 'euclidean')
    total = 0
    for line in dist:
        total += np.sort(line)[1:1 + neighbours].sum() / (neighbours + 1)
    return total / dist.shape[0]


def average_spacing(data, mean_angular_distance):
    """
    Calculates mean distance between points using mean radius of data points and mean angular distance between them
    :param data: numpy.array;

        shape::

            ([[x1 y1 z1],
              [x2 y2 z2],
               ...
              [xN yN zN]])

    :param mean_angular_distance: (numpy.)float; in radians
    :return: float;
    """
    average_radius = np.mean(np.linalg.norm(data, axis=1)) if not np.isscalar(data) else data
    return average_radius * mean_angular_distance


def remap(x, mapper):
    """
    Function rearranges list of points indices according to indices in mapper.
    Maper contains on the nth place new address of the nth point.

    ::

        mapper = <class 'list'>: [7, 8, 9, 1, 2, 3, 4, 5, 6]
        x = <class 'list'>: [[3, 4, 5], [6, 7, 8], [0, 2, 1]]
        result = <class 'list'>: [[1, 2, 3], [4, 5, 6], [7, 9, 8]]

    Value 3 from `x` was replaced by 3rd (in index meaning) value from `mapper`,
    value 6 from `x` is replaced by 6th value from mapper, etc.

    :param x: numpy.array; faces-like matrix
    :param mapper: numpy.array; transformation map - numpy.array(): new_index_of_point = mapper[old_index_of_point]
    :return: List; faces-like matrix
    """
    return list(map(lambda val: [mapper[val[0]], mapper[val[1]], mapper[val[2]]], x))


def poly_areas(polygons):
    """
    Calculates surface areas of triangles, where `triangles` coordinates of points are in `polygons` variable.

    :param polygons: numpy.array; 3d points
    :return: numpy.array
    """
    polygons = np.array(polygons)
    return 0.5 * np.linalg.norm(np.cross(polygons[:, 1] - polygons[:, 0],
                                         polygons[:, 2] - polygons[:, 0]), axis=1)


def triangle_areas(triangles, points):
    """
    Calculates areas of triangles, where `triangles` indexes of vertices which coordinates are stored in `points`.

    :param triangles: numpy.array; indices of triangulation
    :param points: numpy.array; 3d points
    :return: numpy.array;
    """
    return 0.5 * np.linalg.norm(np.cross(points[triangles[:, 1]] - points[triangles[:, 0]],
                                         points[triangles[:, 2]] - points[triangles[:, 0]]), axis=1)


def calculate_distance_matrix(points1, points2, return_join_vector_matrix=False):
    """
    Function returns distance matrix between two sets of points.

    Return matrix consist of distances in order like foloowing::

        [[D(p1[0], p2[0]), D(p1[0], p2[1]), ..., D(p1[0], p2[n])],
         [D(p1[1], p2[0]), D(p1[1], p2[1]), ..., D(p1[1], p2[n])],
         ...
         [D(p1[m], p2[0]), D(p1[m], p2[1]), ..., D(p1[m], p2[n])]]

    If join vector set to True, normalized join vecetors are defined as vectors in between points
    and positions in matrix are related to their distance in matrix above.

    :param points1: numpy.array;
    :param points2: numpy.array;
    :param return_join_vector_matrix: bool; if True, function also returns normalized distance
                                            vectors useful for dot product during calculation of cos
    :return: Tuple[numpy.array, Union[numpy.array, None]]
    """
    # pairwise distance vector matrix
    distance_vector_matrix = operations.create_distance_vector_matrix(points1, points2)

    distance_matrix = operations.calculate_lengths_in_3d_array(distance_vector_matrix)

    if return_join_vector_matrix:
        normalized_distance_vectors = \
            operations.divide_points_in_array_by_constants(distance_vector_matrix, distance_matrix)
        return distance_matrix, normalized_distance_vectors
    else:
        return distance_matrix, None


def find_face_centres(faces):
    """
    Function calculates centres (center of mass) of each supplied face.

    i-th coordinate of center of mas for one given triangle is computed as::

        x_i = (X_j[0] + X_j[1] + X_j[2]) / 3; j, i = [0, 1, 2],

    where X_j is coordinate if j-th corner of face.

    :param faces: numpy.array;

    shape::

        numpy.array([[[x11,y11,z11],
                      [x11,y11,z11],
                      [x12,y12,z12]]
                     [[x21,y21,z21],
                      [x21,y21,z21],
                      [x22,y22,z22]]
                      ...
                      ])

    :return: numpy.array
    """
    return np.mean(faces, axis=1)


def check_missing_kwargs(mandatory_kwargs, supplied_kwargs, instance_of):
    """
    Checks if all `kwargs` are all in `instance kwargs`.
    If missing raise ValuerError with missing `kwargs`.

    :param mandatory_kwargs: List[str]
    :param supplied_kwargs: List[str]
    :param instance_of: class
    :return:
    """
    missing_kwargs = [f"`{kwarg}`" for kwarg in mandatory_kwargs if kwarg not in supplied_kwargs]
    if len(missing_kwargs) > 0:
        raise ValueError(f'Missing argument(s): {", ".join(missing_kwargs)} in class instance {instance_of.__name__}')


def check_missing_params(mandatory_kwargs, supplied_kwargs, instance):
    """
    Checks if all `kwargs` are all in parameter `obj` .
    If missing raise ValuerError with missing `kwargs`.

    :param mandatory_kwargs: List[str]
    :param supplied_kwargs: List[str]
    :param instance: class instance
    :return:
    """
    missing_kwargs = [f"`{kwarg}`" for kwarg in mandatory_kwargs if kwarg not in supplied_kwargs]
    if len(missing_kwargs) > 0:
        raise ValueError(f'Missing argument(s): {", ".join(missing_kwargs)} in object {instance}')


def numeric_logg_to_string(logg):
    """
    Convert numeric form of cgs logg to string used in castelli-kurucz 04,
    kurucz 93 and van-hamme limb darkening table names.

    :param logg: float;
    :return: str;
    """
    return "g%02d" % (logg * 10)


def numeric_metallicity_to_string(metallicity):
    """
    Convert numeric form of cgs metallicity (M/H units) to string used in castelli-kurucz 04,
    kurucz 93 and van-hamme limb darkening table names.

    :param metallicity: float
    :return: str
    """
    sign = "p" if metallicity >= 0 else "m"
    leadzeronum = "%02d" % (metallicity * 10) if metallicity >= 0 else "%02d" % (metallicity * -10)
    return "{sign}{leadzeronum}".format(sign=sign, leadzeronum=leadzeronum)


def numeric_metallicity_from_string(n_metallicity):
    """
    Return numeric metallicity from string used in van-hamme tables.

    :param n_metallicity: str;
    :return: float;
    """
    m = n_metallicity
    sign = 1 if str(m).startswith("p") else -1
    value = float(m[1:]) / 10.0
    return value * sign


def find_nearest_value_as_matrix(look_in, look_for):
    """
    Finds values and indices of elements in `look_in` that are the closest to the each value in `values`.

    :param look_in: numpy.array; elemnts we are looking closest point in
    :param look_for: numpy.array; elements according to which the closest element in `look_in` is searched for
    :return: Tuple[numpy.array, numpy.array]
    """
    val = np.array([look_for]) if np.isscalar(look_for) else look_for
    dif = up.abs(val[:, np.newaxis] - look_in)
    argmins = dif.argmin(axis=1)
    val = look_in[argmins]
    return val, argmins


def find_nearest_value(look_in, look_for):
    """
    Find nearest value in `look_in` array to value `look_for`.

    :param look_in: numpy.array;
    :param look_for: float;
    :return: List[look_for: float, int (index from look_for)];
    """
    look_in = np.array(look_in)
    look_for = look_in[(up.abs(look_in - look_for)).argmin()]
    index = up.where(look_in == look_for)[0][0]
    return [look_for, index]


def find_surrounded_as_matrix(look_in, look_for):
    """
    Find values from `look_in` which souround values from `look_for`. If any value of `look_in` array is exact same
    as value in `look_for` surounded value is same value from left and right side.

    :param look_in: numpy.array;
    :param look_for: numpy.array;
    :return: numpy.array;
    """
    if not (np.array(look_in.min() <= look_for).all() and np.array(look_for <= look_in.max()).all()):
        raise ValueError("At least one value in `look_for` is out of bound of `look_in`")

    dif = look_for[:, np.newaxis] - look_in
    positive_mask = dif >= 0
    # for values on the left side of look_in array
    all_positive = np.all(positive_mask, axis=1)
    # add artificial sign change for right boundary value
    # switch 'fancy' indexing to integer index since in numpy, combined assigment can't be done by fancy indexing)
    all_positive_inline = up.arange(0, len(look_for))[all_positive]
    positive_mask[all_positive_inline, -1] = False
    # find signs switching columns
    sign_swith_mask = up.logical_xor(positive_mask[:, :-1], positive_mask[:, 1:])
    idx_array = np.ones(np.shape(dif), dtype=INT) * up.arange(np.shape(look_in)[0])
    idx_array = idx_array[:, :-1][sign_swith_mask]
    ret_matrix = np.column_stack((look_in[idx_array], look_in[idx_array + 1]))
    # consider on place value as not surounded (surounded by itself)
    isin_look_in = np.isin(look_for, look_in)
    ret_matrix[isin_look_in] = np.array([look_for, look_for]).T[isin_look_in]
    return ret_matrix


def find_surrounded(look_in, look_for):
    """
    Find values from `look_in` which surround `look_for` value.
    If exact `look_for` value is supplied as exists in `look_for` just this one value is returned
    in array as left and right border.

    :param look_in: numpy.array;
    :param look_for: float;
    :return: List [float, float];
    """
    if look_for > max(look_in) or look_for < min(look_in):
        raise ValueError("Any value in `look_for` is out of bound of `look_in`")

    # fixme: just quick hack, make this function better
    if look_for == min(look_in):
        return [min(look_in), min(look_in)]
    elif look_for == max(look_in):
        return [max(look_in), max(look_in)]

    arr, ret = np.array(look_in[:]), []
    f_nst = find_nearest_value(arr, look_for)
    ret.append(f_nst[0])

    new_arr = []
    if f_nst[0] > look_for:
        for i in range(0, len(arr)):
            if arr[i] < f_nst[0]:
                new_arr.append(arr[i])
    else:
        for i in range(0, len(arr)):
            if arr[i] > f_nst[0]:
                new_arr.append(arr[i])

    arr = new_arr[:]
    del new_arr

    # arr = np.delete(arr, f_nst[1], 0)
    ret.append(find_nearest_value(arr, look_for)[0])
    ret = sorted(ret)
    # test
    return ret if ret[0] < look_for < ret[1] else [look_for, look_for]


def calculate_cos_theta(normals, line_of_sight_vector):
    """
    Calculates cosine between two set of normalized vectors::

        - matrix(N * 3), matrix(3) - cosine between each matrix(N * 3) and matrix(3)
        - matrix(N * 3), matrix(M * 3) - cosine between each combination of matrix(N * 3) and matrix(M * 3)

    Combinations are made like normals[0] with all of line_of_sight_vector,
    normals[1] with all from line_of_sight_vector, and so on, and so forth.

    :param normals: numpy.array;
    :param line_of_sight_vector: numpy.array;
    :return: numpy.array;
    """
    line_of_sight_vector = np.array(line_of_sight_vector)
    return np.sum(up.multiply(normals, line_of_sight_vector[None, :]), axis=1) \
        if np.ndim(line_of_sight_vector) == 1 \
        else np.sum(up.multiply(normals[:, None, :], line_of_sight_vector[None, :, :]), axis=2)


def calculate_cos_theta_los_x(normals):
    """
    Calculates cosine of an angle between normalized vectors and line of sight vector [1, 0, 0]

    :param normals: numpy.array
    :return: numpy.array
    """
    return const.LINE_OF_SIGHT[0] * normals[:, 0]


def get_line_of_sight_single_system(phase, inclination):
    """
    Returns line of sight vector for given phase, inclination of the system
    and period of the rotation of given system.

    :param phase: numpy.array;
    :param inclination: float;
    :return: numpy.array;
    """
    line_of_sight_spherical = np.empty((len(phase), 3), dtype=FLOAT)
    line_of_sight_spherical[:, 0] = 1
    line_of_sight_spherical[:, 1] = const.FULL_ARC * phase
    line_of_sight_spherical[:, 2] = inclination
    return spherical_to_cartesian(line_of_sight_spherical)


def convert_gravity_acceleration_array(colormap, units):
    """
    Function converts gravity acceleration array from log_g(SI) units to other units such as `log_cgs`, `SI`, `cgs`.

    :param colormap: numpy.array;
    :param units: str; - `log_cgs`, `SI`, `cgs`, `log_SI`
    :return: numpy.array;
    """
    if units == 'log_cgs':
        # keep it in this way to avoid mutable rewrite of origin colormap array
        # for more information read python docs about __add__ and __iadd__ class methods
        colormap = colormap + 2
    elif units == 'SI':
        colormap = up.power(10, colormap)
    elif units == 'cgs':
        colormap = up.power(10, colormap + 2)
    elif units == 'log_SI':
        pass
    return colormap


def cosine_similarity(a, b):
    """
    Function calculates cosine of angle between vectors.

    :note: Use only in case that a, and b are not normalized, otherwise
           use function calculate_cos_theta; it is way faster since it doesn't normalize vectors on fly.

    :param a: numpy.array;
    :param b: numpy.array;
    :return: float;
    """
    return up.inner(a, b) / (norm(a) * norm(b))


def is_empty(value):
    """
    Consider `value` as empty if fit following rules::

        - is None
        - is Sized with zero length
        - is pandas.DataFrame with zero length
        - is pandas.NaT
        - is numpy.nan

    :param value: Any
    :return: bool
    """
    if isinstance(value, type(None)):
        return True
    if isinstance(value, Sized):
        # this cover also strings
        return len(value) == 0 or len(list(flatten(value))) == 0
    if isinstance(value, pd.DataFrame):
        return value.empty
    if isinstance(value, type(pd.NaT)):
        return True
    if up.isnan(value):
        return True
    return False


def find_idx_of_nearest(array, values):
    """
    Find indices of elements in `array` that are closest to elements in `values`.

    :param array: 1D array (M) - points to be searched for the closest
    :param values: 1D array (N) - values to which closest point in the `array` should be found
    :return: np.array with shape (N) that points to the closest values in `array`
    """
    array = np.asarray(array)
    idx = (up.abs(array[np.newaxis, :] - values[:, np.newaxis])).argmin(axis=1)
    return idx


def rotation_in_spherical(phi, theta, phi_rotation, theta_rotation):
    """
    transformation of phi, theta spherical coordinates into new spherical coordinates produced by rotation around old
    z_axis (in positive direction) by `phi_rotation` and second rotation (in positive direction) around new y axis by
    value `theta rotation`.

    :param phi: numpy.array; - in radians
    :param theta: numpy.array; - in radians
    :param phi_rotation: float; - rotation of old spherical system around z axis, in radians
    :param theta_rotation: float; - rotation of z axis along new y axis by this value, in radians
    :return: Tuple; transformed angular coordinates
    """
    # rotation around Z axis
    phi_rot = (phi - phi_rotation) % const.FULL_ARC

    # rotation around Y axis by `theta_rotation` angle
    cos_phi = up.cos(phi_rot)
    sin_theta = up.sin(theta)
    sin_axis_theta = up.sin(theta_rotation)
    cos_theta = up.cos(theta)
    cos_axis_theta = up.cos(theta_rotation)
    theta_new = up.arccos(cos_phi * sin_theta * sin_axis_theta + cos_theta * cos_axis_theta)
    phi_new = up.arctan2(up.sin(phi_rot) * sin_theta, cos_phi * sin_theta * cos_axis_theta -
                         cos_theta * sin_axis_theta)
    return phi_new, theta_new


def derotation_in_spherical(phi, theta, phi_rotation, theta_rotation):
    """
    backward transformation of spherical coordinates produced by rotation around old
    z_axis by `phi_rotation` and second rotation around new y axis by value `theta rotation` into original coordinate
    system

    :param phi: numpy.array; - in radians
    :param theta: numpy.array; - in radians
    :param phi_rotation: float; - rotation of old spherical system around z axis, in radians
    :param theta_rotation: float; - rotation of z axis along new y axis by this value, in radians
    :return:
    """
    cos_theta = up.cos(theta)
    sin_theta = up.sin(theta)
    cos_phi = up.cos(phi)
    sin_phi = up.sin(phi)

    sin_axis_theta = up.sin(theta_rotation)
    cos_axis_theta = up.cos(theta_rotation)

    theta_new = up.arccos(np.round(cos_theta * cos_axis_theta - cos_phi * sin_theta * sin_axis_theta, 10))
    phi_new = up.arctan2(sin_phi * sin_theta, cos_phi * sin_theta * cos_axis_theta +
                         cos_theta * sin_axis_theta)
    return (phi_new + phi_rotation) % const.FULL_ARC, theta_new


def calculate_equiv_radius(volume):
    """returns equivalent radius of a sphere with given volume"""
    return up.power(3.0 * volume / (4.0 * const.PI), 1.0 / 3.0)


def calculate_ellipsoid_volume(_a, _b, _c):
    """
    Calculates volume of ellipsoid with semi-axis _a, _b and _c.
    :param _a: Union[float, numpy.array];
    :param _b: Union[float, numpy.array];
    :param _c: Union[float, numpy.array];
    :return: Union[float, numpy.array];
    """
    return 4.0 * const.PI * _a * _b * _c / 3.0


class IterableQueue(object):
    """ Transform standard python Queue instance to iterable one"""

    def __init__(self, source_queue):
        """
        :param source_queue: queue.Queue, (mandatory)
        """
        self.source_queue = source_queue

    def __iter__(self):
        while True:
            try:
                yield self.source_queue.get_nowait()
            except Empty:
                return


def is_even(x):
    return x % 2 == 0


def convert_binary_orbital_motion_arr_to_positions(arr):
    return [const.Position(*[int(p[0]) if not np.isnan(p[0]) else p[0]] + list(p[1:])) for p in arr]


def nested_dict_values(dictionary):
    for value in dictionary.values():
        if isinstance(value, dict):
            yield from nested_dict_values(value)
        else:
            yield value


def calculate_volume_ellipse_approx(equator_points=None, meridian_points=None):
    """
    Function calculates volume of the object where only equator and meridian points where provided usin elipsoidal
    approximation for the points with the same x-cordinates.

    :param equator_points: numpy array; (yzx) column-wise
    :param meridian_points: numpy array; (yzx) column-wise
    :return: float;
    """
    areas = up.abs(const.PI * equator_points[:, 1] * meridian_points[:, 0])
    return up.abs(np.trapz(areas, equator_points[:, 2]))


def plane_projection(points, plane, keep_3d=False):
    """
    Function projects 3D points into given plane.

    :param keep_3d: bool; if True, the dimensions of the array is kept the same, with given column equal to zero
    :param points: numpy.array;
    :param plane: str; one of 'xy', 'yz' or 'zx'
    :return: numpy.array;
    """
    rm_index = {"xy": 2, "yz": 0, "zx": 1}[plane]
    if not keep_3d:
        indices_to_keep = [0, 1, 2]
        del indices_to_keep[rm_index]
        return points[:, indices_to_keep]
    in_plane = deepcopy(points)
    in_plane[:, rm_index] = 0.0
    return in_plane


def get_visible_projection(obj):
    """
    Returns yz projection of nearside points.

    :param obj: instance;
    :return: numpy.array
    """
    return plane_projection(
        obj.points[
            np.unique(obj.faces[obj.indices])
        ], "yz"
    )


def split_to_batches(array, n_proc):
    """
    Split array to batches with size `batch_size`.

    :param n_proc: int; number of processes
    :param array: Union[List, numpy.array];
    :return: List;
    """
    indices = np.linspace(0, len(array), num=n_proc+1, endpoint=True, dtype=INT)
    indices = [(indices[ii-1], indices[ii]) for ii in range(1, n_proc+1)]
    return [array[idx[0]: idx[1]] for idx in indices]


def renormalize_async_result(result):
    """
    Renormalize multiprocessing output to native form.
    Multiprocessing will return several dicts with same passband (due to supplied batches), but continuous
    computaion require Dict in form like::

        [{'passband': [all fluxes]}]

    instead::

        [[{'passband': [fluxes in batch]}], [{'passband': [fluxes in batch]}], ...]

    :param result: List;
    :return: Dict[str; numpy.array]
    """
    return {key: np.concatenate([record[key] for record in result], axis=0) for key in result[-1].keys()}


def random_sign():
    """
    Return random sign (-1 or 1)
    """
    random = np.random.randint(0, 2)
    return 1 if random else -1


def str_repalce(x, old, new):
    """
    Replace old values with new in strin `x`.

    :param x: str;
    :param old: Union[str, Iterable[str]];
    :param new: Union[str, Iterable[str]];
    :return: str;
    """
    old = [old] if isinstance(old, str) else old
    new = [new] if isinstance(new, str) else new

    for _old, _new in zip(old, new):
        x = str(x).replace(str(_old), str(_new))
    return x


def magnitude_to_flux(data, zero_point):
    return np.power(10, (zero_point - data) / 2.5)


def magnitude_error_to_flux_error(error):
    return np.power(10, error / 2.5) - 1.0


def flux_to_magnitude(data, zero_point):
    return -2.5*np.log10(data) + zero_point


def flux_error_to_magnitude_error(data, error):
    return 2.5 * np.log10(1 + (error / data))


def discretization_correction_factor(discretization_factor, correction_factors):
    """
    Correction factor for the surface due to underestimation of the surface by the triangles.

    :param correction_factors: numpy.array; (2*N) [discretization factor, correction factor],
                                            sorted according to discretization factor
    :param discretization_factor: numpy.float;
    :return: float;
    """
    # treating edge cases
    if discretization_factor <= correction_factors[0, 0]:
        correction_factor = correction_factors[1, 0]
    elif discretization_factor >= correction_factors[0, -1]:
        correction_factor = correction_factors[1, -1]
    else:
        correction_factor = np.interp(discretization_factor,
                                      correction_factors[0],
                                      correction_factors[1])
    # correction for non-equilateral triangles
    alpha = correction_factor * discretization_factor
    # correction for surface underestimation
    return np.sqrt(alpha / np.sin(alpha))


def transform_values(value, default_unit, unit):
    """
    Quick function for transformation to desired units.

    :param value: Union[float, numpy.array]; input values in default unit
    :param default_unit: astropy.units.Unit; base unit in which `value` is stored
    :param unit: astropy.units.Unit; target unit
    :return: Union[float, numpy.array]; transformed values
    """
    return value if unit == 'default' else (value*default_unit).to(unit).value


def jd_to_phase(times, period, t0, centre=0.5):
    """
    Converting JD to phase according to supplied ephemeris.
    Phases will be returned in range ('centre' - 0.5, 'centre' + 0.5).

    :param times: numpy.array;
    :param period: float;
    :param t0: float;
    :param centre: float;
    :return: numpy.array; converted phases
    """
    start_phase = centre - 0.5
    t0 += start_phase * period
    return ((times - t0) / period) % 1.0 + start_phase


def jd_from_datetime(dt: datetime.datetime):
    """
    Convert a Python datetime object to the Julian Date.

    Parameters:
      dt (datetime): The datetime object to convert. Assumes the Gregorian calendar.

    Returns:
      float: The Julian Date.
    """

    dt_utc = dt.astimezone(datetime.timezone.utc)

    # Extract year, month, and day with the fractional part of the day.
    year = dt_utc.year
    month = dt_utc.month
    # Compute the fractional day: hours, minutes, seconds, and microseconds
    day_fraction = dt_utc.day + (
            dt_utc.hour / 24.0 +
            dt_utc.minute / 1440.0 +
            dt_utc.second / 86400.0 +
            dt_utc.microsecond / 86400.0 / 1e6
    )

    # For January and February, treat them as months 13 and 14 of the previous year
    if month <= 2:
        year -= 1
        month += 12

    # Compute the Gregorian calendar correction:
    a_ = np.floor(year / 100)
    b_ = 2 - a_ + np.floor(a_ / 4)

    # Now, apply the formula.
    jd = (
            np.floor(365.25 * (year + 4716)) +
            np.floor(30.6001 * (month + 1)) +
            day_fraction + b_ - 1524.5
    )

    return jd



import numpy as np

from elisa import u, settings, opt, BinarySystem
from elisa import umpy as up
from elisa.binary_system.model import (
    pre_calculate_for_potential_value_primary,
    radial_primary_potential_derivative,
    potential_primary_fn)
from elisa.binary_system.surface.gravity import calculate_potential_gradient
from elisa.const import Position, HALF_PI, FULL_ARC
from elisa import utils
from elisa.binary_system.container import OrbitalPositionContainer
from elisa.binary_system.surface.coverage import get_eclipse_boundary_path
from elisa.pypex import Polygon


LINE_OF_SIGHT = np.array([1, 0, 0])
BINARY_DEFINITION = {
    "system": {
        "argument_of_periastron": 90.0,
        "gamma": 0.0,
        "period": 5.0,
        "eccentricity": 0.0,
        "inclination": 95.0,
        "primary_minimum_time": 0.0,
        "phase_shift": 0.0
    },
    "primary": {
        "mass": 3.0,
        "surface_potential": 4.2,
        "synchronicity": 1.0,
        "t_eff": 6000.0,
        "gravity_darkening": 0.09,
        "albedo": 0.5,
        "metallicity": 0.0,
        "discretization_factor": 5
    },
    "secondary": {
        "mass": 0.5,
        "surface_potential": 5.0,
        "synchronicity": 1.0,
        "t_eff": 5000.0,
        "gravity_darkening": 0.09,
        "albedo": 0.5,
        "metallicity": 0.0
    }
}


def _horizon_base_component(binary, position, analytic=True):
    if analytic:
        binary.primary.discretization_factor = 1.0 * u.deg
        binary.secondary.discretization_factor = 10.0 * u.deg
        binary.init()

    position_container = OrbitalPositionContainer.from_binary_system(binary, position)
    position_container.build_mesh(components_distance=1.0, component="primary")
    position_container.build_faces(components_distance=1.0, component="primary")
    position_container.build_faces_orientation(components_distance=1.0, component="primary")
    position_container.build_surface_areas(component="primary")

    if not analytic:
        # rotate
        for prop in ["points", "normals"]:
            prop_value = getattr(position_container.primary, prop)

            args = (position_container.position.azimuth - HALF_PI, prop_value, "z", False, False)
            prop_value = utils.around_axis_rotation(*args)

            args = (HALF_PI - position_container.inclination, prop_value, "y", False, False)
            prop_value = utils.around_axis_rotation(*args)
            setattr(position_container.primary, prop, prop_value)

        # compute los cosines
        normals = getattr(position_container.primary, "normals")
        los_cosines = normals[:, 0]
        setattr(position_container.primary, "los_cosines", los_cosines)

        # apply darkside filter (horizon)
        cosines = getattr(position_container.primary, "los_cosines")
        valid_indices = position_container.darkside_filter(cosines=cosines)
        setattr(position_container.primary, "indices", valid_indices)

    return position_container


def estimate_analytic_horizon(binary=None, phase=0.0, threshold=-1e-6, polar=False, cosine_precision=False, _3d=False):
    """
    Estimate analytic horizon of primary component.
    This use very dense discretization of surface to approach real horizon
    without solving complicated equations.
    """

    def _cosine_precision(cosine):
        precisions = reversed(1.0 / np.power(10, [x for x in range(1, 11)]))
        for _precision in precisions:
            if np.sum(cosine > _precision) == 0:
                return _precision
        return -1

    if binary is None:
        binary = BinarySystem.from_json(BINARY_DEFINITION)
    position = Position(*((0,) + tuple(binary.orbit.orbital_motion(phase=phase)[0])))
    container = _horizon_base_component(binary, position, analytic=True)
    star = container.primary
    normals = calculate_potential_gradient(position.distance, "primary", star.points,
                                           star.synchronicity, binary.mass_ratio)
    properties = []
    for props in [star.points, normals]:
        args = (position.azimuth - HALF_PI, props, "z", False, False)
        prop_value = utils.around_axis_rotation(*args)
        args = (HALF_PI - container.inclination, prop_value, "y", False, False)
        prop_value = utils.around_axis_rotation(*args)
        properties.append(prop_value)

    del container

    points, normals = properties
    cosines = np.inner(normals, LINE_OF_SIGHT)
    valid_indices = OrbitalPositionContainer.darkside_filter(cosines)

    visible_projection = utils.plane_projection(points[valid_indices], "yz")
    bb_path = get_eclipse_boundary_path(visible_projection)
    horizon_indices = up.invert(bb_path.contains_points(visible_projection, radius=threshold))

    precision = None
    if cosine_precision:
        precision = _cosine_precision(cosines[valid_indices][horizon_indices])

    if _3d:
        return points[valid_indices][horizon_indices]

    horizon = visible_projection[horizon_indices]

    if polar:
        horizon = utils.cartesian_to_polar(horizon)
        horizon_argsort = np.argsort(horizon.T[1])
        horizon = horizon[horizon_argsort]

    return (horizon, precision) if cosine_precision else horizon


def get_analytics_horizon(binary=None, phase=0.0, tol=1e-4, polar=False, phi_density=100, theta_density=1000):
    if binary is None:
        binary = BinarySystem.from_json(BINARY_DEFINITION)

    star = binary.primary
    position = Position(*((0,) + tuple(binary.orbit.orbital_motion(phase=phase)[0])))

    # rotate line of sight to simulate phase and inclination
    zv = np.array([0.0, 0.0, 1.0])

    xv = utils.around_axis_rotation(HALF_PI - binary.inclination, LINE_OF_SIGHT, axis="y", inverse=True)
    xv = utils.around_axis_rotation(position.azimuth - HALF_PI, xv, axis="z", inverse=True)

    zv = utils.around_axis_rotation(HALF_PI - binary.inclination, zv, axis="y", inverse=True)
    zv = utils.around_axis_rotation(position.azimuth - HALF_PI, zv, axis="z", inverse=True)

    # perpendicular vector to find theta-like rotation
    yv = np.cross(zv, xv)

    potential_fn = potential_primary_fn
    precalc_fn = pre_calculate_for_potential_value_primary
    fprime = radial_primary_potential_derivative

    # prepare-phi like vector
    phi_range = np.linspace(np.radians(0), np.radians(360), phi_density)
    theta_range = np.linspace(np.radians(-5), np.radians(5), theta_density)

    # prepare theta-like vectors via rotation around phi and then aroun yv in -/+ range
    vectors = []
    for d_phi in phi_range:
        # girst rotate zv about phi around vector `xv`
        inner_vectors = []
        vector = utils.arbitrary_rotation(d_phi, xv, vector=zv)
        _yv = utils.arbitrary_rotation(d_phi, xv, vector=yv)
        for d_theta in theta_range:
            vectors.append(utils.arbitrary_rotation(d_theta, _yv, vector=vector))

    vectors = np.array(utils.cartesian_to_spherical(vectors))

    phi, theta = vectors[:, 1], vectors[:, 2]
    args = (star.synchronicity, binary.mass_ratio, position.distance, phi, theta)
    precalc_vals = precalc_fn(*args, return_as_tuple=True)
    x0 = star.side_radius * np.ones(phi.shape)
    radius_kwargs = dict(fprime=fprime, maxiter=settings.MAX_SOLVER_ITERS, rtol=1e-10,
                         args=((binary.mass_ratio,) + precalc_vals, star.surface_potential))

    radius = opt.newton.newton(potential_fn, x0, **radius_kwargs)
    points = utils.spherical_to_cartesian(np.array([radius, phi, theta]).T)
    normals = calculate_potential_gradient(1.0, "primary", points, star.synchronicity, binary.mass_ratio)

    cosines = np.inner(normals, xv)
    cosines = cosines.reshape(phi_density, theta_density)
    cosines_gtz = [up.arange(np.shape(row)[0])[row > 0] for row in cosines]
    # take only smallest values (but greater than zero) in each theta line
    cosines_argmin = [np.argmin(row[gtz]) for row, gtz in zip(cosines, cosines_gtz)]

    # find cosines in tolerance (tol)
    valid_argmin = [tol > row[gtz][argmin]
                    for row, gtz, argmin in zip(cosines, cosines_gtz, cosines_argmin)]

    points = points.reshape(phi_density, theta_density, 3)
    horizon_points = [row[gtz][argmin] for row, gtz, argmin, valid in
                      zip(points, cosines_gtz, cosines_argmin, valid_argmin) if valid]

    if utils.is_empty(horizon_points):
        raise ValueError(f"No horizon points found in given tolerance {tol}. Decrease tolerance.")

    horizon_points = utils.around_axis_rotation(position.azimuth - HALF_PI, horizon_points, axis="z", inverse=False)
    horizon_points = utils.around_axis_rotation(HALF_PI - binary.inclination, horizon_points, axis="y", inverse=False)
    horizon_points = horizon_points.T[1:3].T

    if polar:
        horizon_points = utils.cartesian_to_polar(horizon_points)
        horizon_argsort = np.argsort(horizon_points.T[1])
        horizon_points = horizon_points[horizon_argsort]

    return horizon_points


def _cover_horizon_edges(horizon):
    horizon_polygon = Polygon(horizon)
    horizon = list()
    for edge in horizon_polygon.edges(as_line=True):
        parametrized = edge.parametrized()
        for t in np.arange(0, 1, 0.01):
            horizon.append(parametrized(t))
    return np.array(horizon)


def get_discrete_horizon(binary=None, phase=0.0, threshold=-1e-6, polar=False):
    """
    Get discrete horizon of primary component.
    """
    if binary is None:
        binary = BinarySystem.from_json(BINARY_DEFINITION)

    position = Position(*((0,) + tuple(binary.orbit.orbital_motion(phase=phase)[0])))
    position_container = _horizon_base_component(binary, position, analytic=False)
    position_container.correct_mesh(component="primary", components_distance=position_container.position.distance)
    star = position_container.primary
    visible_projection = utils.get_visible_projection(star)

    bb_path = get_eclipse_boundary_path(visible_projection)
    horizon_indices = up.invert(bb_path.contains_points(visible_projection, radius=threshold))
    horizon = visible_projection[horizon_indices]
    origin_horizon = horizon
    horizon = _cover_horizon_edges(horizon)

    if polar:
        horizon = utils.cartesian_to_polar(horizon)
        horizon_argsort = np.argsort(horizon.T[1])

        origin_horizon = utils.cartesian_to_polar(origin_horizon)
        origin_horizon_argsort = np.argsort(origin_horizon.T[1])
        return horizon[horizon_argsort], origin_horizon[origin_horizon_argsort]

    return horizon, origin_horizon


if __name__ == "__main__":
    from matplotlib import pyplot as plt
    _phase = 0.25

    discrete_horizon, origin_discrete_horizon = get_discrete_horizon(phase=_phase, polar=False)

    # show full path of discrete horizon
    # phi_argsort = np.argsort(discrete_horizon.T[1] % FULL_ARC)
    # rs, phis = discrete_horizon[phi_argsort].T[0], discrete_horizon[phi_argsort].T[1] % FULL_ARC
    # rs, phis = rs[:-1], phis[:-1]
    # plt.plot(phis % FULL_ARC, rs * 10, c="r")

    # analytic_horizon = get_analytics_horizon(phase=_phase, tol=1e-2, polar=False, phi_density=50, theta_density=1000)
    #
    # plt.scatter(analytic_horizon.T[0], analytic_horizon.T[1], c="r")
    # plt.show()

    # # show vertex path of discrete horizon
    # phi_argsort = np.argsort(origin_discrete_horizon.T[1] % FULL_ARC)
    # rs, phis = origin_discrete_horizon[phi_argsort].T[0], origin_discrete_horizon[phi_argsort].T[1] % FULL_ARC
    # rs, phis = rs[:-1], phis[:-1]
    #
    # plt.scatter(phis % FULL_ARC, rs * 10, c="r")
    #
    # # analytic horizon
    # analytic_horizon = get_analytics_horizon(phase=_phase, tol=1e-2, polar=True, phi_density=100, theta_density=1000)
    # phi_argsort = np.argsort(analytic_horizon.T[1] % FULL_ARC)
    # rs, phis = analytic_horizon[phi_argsort].T[0], analytic_horizon[phi_argsort].T[1] % FULL_ARC
    # rs, phis = rs[:-1], phis[:-1]
    #
    # plt.plot(phis % FULL_ARC, rs * 10, c="b")
    # plt.xlabel(r"$\theta$")
    # plt.ylabel(r"$\varrho$")
    # plt.legend()
    # plt.show()



import numpy as np

from .. import (
    utils as bsutils,
    model,
)
from ...base.error import MaxIterationError, SpotError
from ...base.spot import incorporate_spots_mesh
from ...base.surface.mesh import correct_component_mesh
from ... import settings
from ...opt.fsolver import fsolver
from ...utils import is_empty
from ...logger import getLogger
from ... import (
    umpy as up,
    utils,
    opt,
    const
)

logger = getLogger("binary_system.surface.mesh")

CORRECTION_FACTORS = {
    'detached': np.load(settings.PATH_TO_DETACHED_CORRECTIONS, allow_pickle=False),
    'over-contact': np.load(settings.PATH_TO_OVER_CONTACT_CORRECTIONS, allow_pickle=False)
}
CORRECTION_FACTORS['semi-detached'] = CORRECTION_FACTORS['detached']
CORRECTION_FACTORS['double-contact'] = CORRECTION_FACTORS['detached']


def build_mesh(system, components_distance, component="all"):
    """
    Build surface points for primary or/and secondary component. In case of spots,
    the spot point mesh is incorporated into the model. Points are assigned to system.

    :param system: elisa.binary_system.contaier.OrbitalPositionContainer; instance
    :param component: Union[str, None];
    :param components_distance: float;
    :return: system; elisa.binary_system.contaier.OrbitalPositionContainer; instance
    """
    components = bsutils.component_to_list(component)

    for component in components:
        star = getattr(system, component)
        # in case of spoted surface, symmetry is not used
        if getattr(system, 'morphology') == 'over-contact':
            a, c, d = mesh_over_contact(system, component, symmetry_output=True)
        else:
            a, c, d = mesh_detached(system, components_distance, component, symmetry_output=True)

        star.points = a
        star.base_symmetry_points_number = c
        star.inverse_point_symmetry_matrix = d

    add_spots_to_mesh(system, components_distance, component="all")

    return system


def rebuild_symmetric_detached_mesh(system, components_distance, component):
    """
    Rebuild a mesh of a symmetrical surface using old mesh to provide azimuths for the new. This conserved number
    of points and faces.

    :param system: elisa.binary_system.contaier.OrbitalPositionContainer; instance
    :param component: Union[str, None];
    :param components_distance: float;
    :return: system; elisa.binary_system.contaier.OrbitalPositionContainer; instance
    """
    components = bsutils.component_to_list(component)

    for component in components:
        star = getattr(system, component)
        setattr(star, "points", rebuild_mesh_detached(system, components_distance, component))

    return system


def pre_calc_azimuths_for_detached_points(discretization, star):
    """
    Router for discretization method of detached surfaces.

    :param discretization: float; discretization factor
    :param star:
    :return: Tuple; (phi: numpy.array, theta: numpy.array, separator: numpy.array)
    """
    if settings.MESH_GENERATOR in ['auto', 'improved_trapezoidal']:
        rel_radii = (star.forward_radius - star.polar_radius) / star.polar_radius
        if rel_radii > settings.DEFORMATION_TOL or settings.MESH_GENERATOR == 'improved_trapezoidal':
            args = discretization, star.forward_radius, star.polar_radius, star.side_radius, star.backward_radius
            return improved_trapezoidal_mesh(*args)

    return trapezoidal_mesh(discretization)


def trapezoidal_mesh(discretization):
    """
    Caculates azimuths for every surface point using trapezoidal discretization. Good for nearly spherical stars.

    :param discretization: float;
    :return: Tuple;
    """
    # vertical alpha needs to be corrected to maintain similar sizes of triangle sizes
    # vertical_alpha = const.POINT_ROW_SEPARATION_FACTOR * discretization
    vertical_alpha = discretization

    separator = []

    # azimuths for points on equator
    num = int(const.PI // (const.SEAM_CONST * discretization))
    phi = np.linspace(0., const.PI, num=num + 1)
    theta = np.full(phi.shape, const.HALF_PI)
    separator.append(np.shape(theta)[0])

    # azimuths for points on meridian
    v_num = int(const.HALF_PI // (const.SEAM_CONST * vertical_alpha))
    # v_num = int(const.HALF_PI // discretization)
    phi_meridian = np.concatenate((const.PI * np.ones(v_num - 1), np.zeros(v_num)))
    theta_meridian = up.concatenate((np.linspace(const.HALF_PI, 0, num=v_num + 1)[1:-1],
                                     np.linspace(0., const.HALF_PI, num=v_num, endpoint=False)))

    phi = up.concatenate((phi, phi_meridian))
    theta = up.concatenate((theta, theta_meridian))
    separator.append(np.shape(theta)[0])

    v_num = int(const.HALF_PI // vertical_alpha)
    # azimuths for rest of the quarter
    thetas = np.linspace(discretization, const.HALF_PI, num=v_num - 1, endpoint=False)
    for tht in thetas:
        alpha_corrected = discretization / up.sin(tht)
        num = int(const.PI // alpha_corrected)
        alpha_corrected = const.PI / (num + 1)
        phi_q_add = alpha_corrected * np.arange(1, num + 1)
        phi = up.concatenate((phi, phi_q_add))
        theta = up.concatenate((theta, tht * np.ones(phi_q_add.shape[0])))

    return phi, theta, separator


def improved_trapezoidal_mesh(discretization, forward_radius, polar_radius, side_radius, backward_radius):
    """
    Calculates azimuths for every surface point using trapezoidal discretization. Function conserves areas of the
    triangles better than trapezoidal method.

    :param discretization: numpy.float;
    :param forward_radius: numpy.float;
    :param polar_radius: numpy.float;
    :param side_radius: numpy.float;
    :param backward_radius: numpy.float;
    :return: Tuple; (phi: numpy.array, theta: numpy.array, separtor: List)
    """
    # vertical alpha needs to be corrected to maintain similar sizes of triangle sizes
    # vertical_alpha = const.POINT_ROW_SEPARATION_FACTOR * discretization
    vertical_alpha = discretization

    separator = []

    # azimuths for points on equator
    num = int(const.PI // (const.SEAM_CONST * discretization))
    phi = np.linspace(0., const.PI, num=num + 1)
    theta = np.full(phi.shape, const.HALF_PI)
    separator.append(np.shape(theta)[0])
    # obliqueness correction
    inner_mask = (phi < const.HALF_PI) & (phi > 0)
    outer_mask = (phi > const.HALF_PI) & (phi < const.PI)
    inner_phis = phi[inner_mask]
    outer_phis = phi[outer_mask]
    tan_phs1 = up.tan(inner_phis)
    tan_phs2 = up.tan(outer_phis)
    inner_corr = up.arctan((side_radius - forward_radius) * tan_phs1 /
                           (side_radius + forward_radius * tan_phs1 ** 2))
    outer_corr = up.arctan((side_radius - backward_radius) * tan_phs2 /
                           (side_radius + backward_radius * tan_phs2 ** 2))
    phi[inner_mask] += inner_corr
    phi[outer_mask] += outer_corr

    # azimuths for points on meridian
    v_num = int(const.HALF_PI // (1.07 * vertical_alpha))
    # v_num = int(const.HALF_PI // discretization)
    phi_meridian = np.concatenate((const.PI * np.ones(v_num - 1), np.zeros(v_num)))
    theta_meridian = up.concatenate((np.linspace(const.HALF_PI, 0, num=v_num + 1)[1:-1],
                                     np.linspace(0., const.HALF_PI, num=v_num, endpoint=False)))

    # azimuths for rest of the quarter
    v_num = int(const.HALF_PI // vertical_alpha)
    thetas_lin = np.linspace(discretization, const.HALF_PI, num=v_num - 1, endpoint=False)

    # correcting theta for obliqueness
    est_eqt_r = (side_radius + forward_radius + backward_radius) / 3.0
    tan_tht = np.tan(theta_meridian)
    theta_meridian += up.arctan((est_eqt_r - polar_radius) * tan_tht /
                                (polar_radius + est_eqt_r * tan_tht ** 2))

    phi = up.concatenate((phi, phi_meridian))
    theta = up.concatenate((theta, theta_meridian))
    separator.append(np.shape(theta)[0])

    # correcting theta for obliqueness
    tan_tht = np.tan(thetas_lin)
    thetas = thetas_lin + up.arctan((est_eqt_r - polar_radius) * tan_tht /
                                    (polar_radius + est_eqt_r * tan_tht ** 2))

    for ii, tht in enumerate(thetas):
        alpha_corrected = discretization / up.sin(tht)
        num = int(const.PI // alpha_corrected)
        alpha_corrected = const.PI / (num + 1)
        phi_q_add = alpha_corrected * np.arange(1, num + 1)

        # correction for obliqness
        inner_mask = phi_q_add < const.HALF_PI
        outer_mask = phi_q_add > const.HALF_PI
        inner_phis = phi_q_add[inner_mask]
        outer_phis = phi_q_add[outer_mask]

        tan_phs1 = up.tan(inner_phis)
        tan_phs2 = up.tan(outer_phis)
        scaling_factor = np.sin(tht)
        inner_corr = np.arctan(scaling_factor * (side_radius - forward_radius) * tan_phs1 /
                               (side_radius + forward_radius * tan_phs1 ** 2))
        outer_corr = np.arctan(scaling_factor * (side_radius - backward_radius) * tan_phs2 /
                               (side_radius + backward_radius * tan_phs2 ** 2))
        phi_q_add[inner_mask] += inner_corr
        phi_q_add[outer_mask] += outer_corr

        phi = up.concatenate((phi, phi_q_add))
        thetas_add = tht * np.ones(phi_q_add.shape[0])
        theta = up.concatenate((theta, thetas_add))
    return phi, theta, separator


def pre_calc_azimuths_for_overcontact_points(discretization, star, component, neck_position, neck_polynomial):
    """
    Router for farside over-contact discretization methods.

    :param discretization: numpy.float;
    :param star: elisa.base.container.StarContainer;
    :param component: str;;
    :param neck_position: numpy.float;
    :param neck_polynomial: numpy.polynomial.Polynomial;
    :return: Tuple;
    """
    if settings.MESH_GENERATOR in ['auto', 'improved_trapezoidal']:
        rel_radii = (star.backward_radius - star.polar_radius) / star.polar_radius
        if rel_radii > settings.DEFORMATION_TOL or settings.MESH_GENERATOR == 'improved_trapezoidal':
            far_azim = improved_trapezoidal_overcontact_farside_points(discretization, star.polar_radius,
                                                                       star.side_radius, star.backward_radius)
            near_azim = improved_trapezoidal_overcontact_neck_points(discretization, neck_position, neck_polynomial,
                                                                     star.polar_radius, star.side_radius, component)
            return far_azim, near_azim

    far_azim = trapezoidal_overcontact_farside_points(discretization)
    near_azim = trapezoidal_overcontact_neck_points(discretization, neck_position, neck_polynomial, star.polar_radius,
                                                    component)
    return far_azim, near_azim


def trapezoidal_overcontact_farside_points(discretization):
    """
    Calculates azimuths (directions) to the surface points of over-contact component on its far-side (convex part).

    :param discretization: float; discretization factor
    :return: Tuple; (phi: numpy.array, theta: numpy.array, separtor: list)
    """
    # vertical alpha needs to be corrected to maintain similar sizes of triangle sizes
    # vertical_alpha = const.POINT_ROW_SEPARATION_FACTOR * discretization
    vertical_alpha = discretization
    separator = []

    # calculating points on farside equator
    num = int(const.HALF_PI // (const.SEAM_CONST * discretization))
    phi = np.linspace(const.HALF_PI, const.PI, num=num + 2)
    theta = np.full(phi.shape, const.HALF_PI)
    separator.append(np.shape(theta)[0])

    # calculating points on phi = pi meridian
    v_num = int(const.HALF_PI / (const.SEAM_CONST * vertical_alpha))
    phi_meridian1 = np.full(v_num - 1, const.PI)
    theta_meridian1 = np.linspace(0., const.HALF_PI, num=v_num - 1, endpoint=False)
    phi = up.concatenate((phi, phi_meridian1))
    theta = up.concatenate((theta, theta_meridian1))
    separator.append(np.shape(theta)[0])

    # calculating points on phi = pi/2 meridian, perpendicular to component`s distance vector
    v_num -= 1
    phi_meridian2 = np.full(v_num - 1, const.HALF_PI)
    theta_meridian2 = np.linspace(0, const.HALF_PI, num=v_num, endpoint=False)[1:]
    phi = up.concatenate((phi, phi_meridian2))
    theta = up.concatenate((theta, theta_meridian2))
    separator.append(np.shape(theta)[0])

    v_num = int(const.HALF_PI / vertical_alpha)
    theta_meridian = np.linspace(0., const.HALF_PI, num=v_num - 1, endpoint=False)
    # calculating the rest of the surface on farside
    for tht in theta_meridian[1:]:
        alpha_corrected = discretization / up.sin(tht)
        num = int(const.HALF_PI // alpha_corrected)
        alpha_corrected = const.HALF_PI / (num + 1)
        phi_q_add = const.HALF_PI + alpha_corrected * np.arange(1, num + 1)
        phi = np.concatenate((phi, phi_q_add))
        theta = np.concatenate((theta, np.full(phi_q_add.shape, tht)))
    separator.append(np.shape(theta)[0])

    return phi, theta, separator


def improved_trapezoidal_overcontact_farside_points(discretization, polar_radius, side_radius, backward_radius):
    """
    Calculates azimuths (directions) to the surface points of over-contact component on its far-side (convex part)
    using improved trapezoidal mesh approach.

    :param discretization: float; discretization factor
    :param polar_radius: numpy.float;
    :param side_radius: numpy.float;
    :param backward_radius: numpy.float;
    :return: Tuple; (phi: numpy.array, theta: numpy.array, separtor: List)
    """
    # vertical alpha needs to be corrected to maintain similar sizes of triangle sizes
    # vertical_alpha = const.POINT_ROW_SEPARATION_FACTOR * discretization
    vertical_alpha = discretization
    separator = []

    # calculating points on farside equator
    num = int(const.HALF_PI / (const.SEAM_CONST * discretization))
    phi = np.linspace(const.HALF_PI, const.PI, num=num + 2)
    theta = np.full(phi.shape, const.HALF_PI)
    separator.append(np.shape(theta)[0])
    # obliqueness correction
    tan_phs = up.tan(phi)
    corr = up.arctan((side_radius - backward_radius) * tan_phs /
                     (side_radius + backward_radius * tan_phs ** 2))
    phi += corr

    # calculating points on phi = pi meridian
    v_num = int(const.HALF_PI / (const.SEAM_CONST * vertical_alpha))
    phi_meridian1 = np.full(v_num - 1, const.PI)
    theta_meridian1 = np.linspace(0., const.HALF_PI, num=v_num - 1, endpoint=False)
    # obliqueness correction
    est_eqt_r = (side_radius + 2 * backward_radius) / 3.0
    tan_tht = np.tan(theta_meridian1)
    theta_meridian1 += up.arctan((est_eqt_r - polar_radius) * tan_tht /
                                 (polar_radius + est_eqt_r * tan_tht ** 2))

    phi = up.concatenate((phi, phi_meridian1))
    theta = up.concatenate((theta, theta_meridian1))
    separator.append(np.shape(theta)[0])

    # calculating points on phi = pi/2 meridian, perpendicular to component`s distance vector
    v_num -= 1
    phi_meridian2 = np.full(v_num - 1, const.HALF_PI)
    theta_meridian2 = theta_meridian1[1:]
    phi = up.concatenate((phi, phi_meridian2))
    theta = up.concatenate((theta, theta_meridian2))
    separator.append(np.shape(theta)[0])

    v_num = int(const.HALF_PI / vertical_alpha)
    theta_meridian = np.linspace(0., const.HALF_PI, num=v_num - 1, endpoint=False)
    tan_tht = np.tan(theta_meridian)
    theta_meridian += up.arctan((est_eqt_r - polar_radius) * tan_tht /
                                (polar_radius + est_eqt_r * tan_tht ** 2))
    # calculating the rest of the surface on farside
    for tht in theta_meridian[1:]:
        alpha_corrected = discretization / up.sin(tht)
        num = int(const.HALF_PI // alpha_corrected)
        alpha_corrected = const.HALF_PI / (num + 1)
        phi_q_add = const.HALF_PI + alpha_corrected * np.arange(1, num + 1)
        # obliqueness correction
        scaling_factor = np.sin(tht)
        tan_phs = up.tan(phi_q_add)
        corr = np.arctan(scaling_factor * (side_radius - backward_radius) * tan_phs /
                         (side_radius + backward_radius * tan_phs ** 2))
        phi_q_add += corr

        phi = np.concatenate((phi, phi_q_add))
        theta = np.concatenate((theta, np.full(phi_q_add.shape, tht)))
    separator.append(np.shape(theta)[0])

    return phi, theta, separator


def _generate_neck_zs(delta_z, component, neck_position, neck_polynomial):
    """
    Common generating function for azimuths on the neck of over-contact systems.

    :param delta_z: numpy.float;
    :param component: str;
    :param neck_position: numpy.float;
    :param neck_polynomial: numpy.polynomial.Polynomial
    :return: Tuple
    """
    # lets define cylindrical coordinate system r_n, phi_n, z_n for our neck where z_n = x, phi_n = 0 heads along
    # z axis

    # alpha along cylindrical axis z needs to be corrected to maintain similar sizes of triangle sizes
    # delta_z = const.POINT_ROW_SEPARATION_FACTOR * delta_z
    delta_z = delta_z
    delta_z_polar = const.SEAM_CONST * delta_z
    # test radii on neck_position
    separator = []

    if component == 'primary':
        num = 100 * int(neck_position // delta_z)
        # position of z_n adapted to the slope of the neck, gives triangles with more similar areas
        x_curve = np.linspace(0., neck_position, num=num, endpoint=True)
        z_curve = np.polyval(neck_polynomial, x_curve)

        curve = np.column_stack((x_curve, z_curve))
        lengths = up.sqrt(np.sum(np.diff(curve, axis=0) ** 2, axis=1))
        neck_lengths = np.cumsum(lengths)
        num_z = int(neck_lengths[-1] // delta_z)
        num_z_polar = int(neck_lengths[-1] // delta_z_polar)
        segments = np.linspace(0, neck_lengths[-1], num=num_z)[1:]
        segments_polar = np.linspace(0, neck_lengths[-1], num=num_z_polar)[1:]

        z_ns = np.interp(segments, neck_lengths, x_curve[1:])
        z_ns_polar = np.interp(segments_polar, neck_lengths, x_curve[1:])
        r_neck = np.polyval(neck_polynomial, z_ns)
    else:
        num = 100 * int((1 - neck_position) // delta_z)
        # position of z_n adapted to the slope of the neck, gives triangles with more similar areas
        x_curve = np.linspace(neck_position, 1, num=num, endpoint=True)
        z_curve = np.polyval(neck_polynomial, x_curve)

        curve = np.column_stack((x_curve, z_curve))
        lengths = up.sqrt(np.sum(np.diff(curve, axis=0) ** 2, axis=1))
        neck_lengths = np.cumsum(lengths)
        num_z = int(neck_lengths[-1] // delta_z)
        num_z_polar = int(neck_lengths[-1] // delta_z_polar)
        segments = np.linspace(0, neck_lengths[-1], num=num_z)[:-1]
        segments_polar = np.linspace(0, neck_lengths[-1], num=num_z_polar)[:-1]

        z_ns = np.interp(segments, neck_lengths, x_curve[:-1])
        z_ns_polar = np.interp(segments_polar, neck_lengths, x_curve[:-1])
        r_neck = np.polyval(neck_polynomial, z_ns)
        z_ns = 1 - z_ns
        z_ns_polar = 1 - z_ns_polar

    # equator azimuths
    phi = np.full(z_ns_polar.shape, const.HALF_PI)
    z = z_ns_polar
    separator.append(np.shape(z)[0])
    # meridian azimuths
    phi = up.concatenate((phi, np.zeros(z_ns_polar.shape)))
    z = up.concatenate((z, z_ns_polar))
    separator.append(np.shape(z)[0])

    return phi, z, z_ns, r_neck, separator


def trapezoidal_overcontact_neck_points(
        discretization, neck_position, neck_polynomial, polar_radius, component):
    """
    Calculates azimuths (directions) to the surface points of over-contact component on neck.

    :param discretization: float; doscretiozation factor
    :param neck_position: float; x position of neck of over-contact binary
    :param neck_polynomial: scipy.Polynome; polynome that define neck profile in plane `xz`
    :param polar_radius: float;
    :param component: str; `primary` or `secondary`
    :return: Tuple; (phi: numpy.array, z: numpy.array, separator: numpy.array)
    """
    # generating the neck
    delta_z = discretization * polar_radius

    phi, z, z_ns, r_neck, separator = _generate_neck_zs(delta_z, component, neck_position, neck_polynomial)

    for ii, zz in enumerate(z_ns):
        num = int(const.HALF_PI * r_neck[ii] // delta_z)
        num = num + 1 if num < 5 else num
        phis = np.linspace(0, const.HALF_PI, num=int(num), endpoint=False)[1:]
        z_n = np.full(phis.shape, zz)

        phi = np.concatenate((phi, phis))
        z = up.concatenate((z, z_n))

    separator.append(np.shape(z)[0])

    return phi, z, separator


def improved_trapezoidal_overcontact_neck_points(
        discretization, neck_position, neck_polynomial, polar_radius, side_radius, component):
    """
    Calculates azimuths (directions) to the surface points of over-contact component on neck.

    :param discretization: float; doscretization factor
    :param neck_position: float; x position of neck of over-contact binary
    :param neck_polynomial: numpy.polynomial.Polynomial; polynome that define neck profile in plane `xz`
    :param polar_radius: float;
    :param side_radius: float;
    :param component: str; `primary` or `secondary`
    :return: Tuple; (phi: numpy.array, z: numpy.array, separator: numpy.array)
    """
    # generating the neck
    delta_z = discretization * polar_radius

    phi, z, z_ns, r_neck, separator = _generate_neck_zs(delta_z, component, neck_position, neck_polynomial)

    eq_coeff = side_radius / polar_radius
    for ii, zz in enumerate(z_ns):
        num = const.HALF_PI * r_neck[ii] // delta_z
        num = num + 1 if num < 4 else num
        phis = np.linspace(0, const.HALF_PI, num=int(num), endpoint=False)[1:]
        # obliqueness correction
        tan_phis = np.tan(phis)
        phis += up.arctan((eq_coeff - 1) * tan_phis /
                          (1 + eq_coeff * tan_phis ** 2))

        z_n = np.full(phis.shape, zz)

        phi = np.concatenate((phi, phis))
        z = up.concatenate((z, z_n))

    separator.append(np.shape(z)[0])

    return phi, z, separator


def get_surface_points(*args):
    """
    Function solves radius for given azimuths that are passed in `args`.
    It use `scipy.optimize.fsolve` method. Function to solve is specified as last parameter in `args` Tuple.
    :param args: Tuple;

    ::

        Tuple[
                phi: numpy.array,
                theta: numpy.array,
                x0: float,
                components_distance: float,
                precalc_fn: callable,
                potential_fn: callable,
                fprime: callable,
                surface_potential: float,
                mass_ratio: float
                synchronicity: float
            ]

    :return: numpy.array
    """
    phi, theta, x0, components_distance, precalc_fn, potential_fn, fprime, potential, q, synchronicity = args
    max_iter = settings.MAX_SOLVER_ITERS
    precalc_vals = precalc_fn(*(synchronicity, q, components_distance, phi, theta), return_as_tuple=True)
    x0 *= np.ones(phi.shape)
    radius_kwargs = dict(fprime=fprime, maxiter=max_iter, args=((q,) + precalc_vals, potential), rtol=1e-10)
    radius = opt.newton.newton(potential_fn, x0, **radius_kwargs)
    if (radius < 0.0).any():
        raise ValueError('Solver found at least one point in the opposite direction. Check you points. ')
    return utils.spherical_to_cartesian(np.column_stack((radius, phi, theta)))


def get_surface_points_cylindrical(*args):
    """
    Function solves radius for given azimuths that are passed in `args`.

    :param args: Tuple;

    ::

         Tuple[
                phi: numpy.array,
                z: numpy.array,
                components_distance: float,
                x0: float,
                precalc_fn: callable,
                potential_fn: callable,
                fprime: callable (fprime),
                surface_potential: float,
                mass_ratio: float,
                synchronicity: float
              ]

    :return: numpy.array;
    """
    phi, z, components_distance, x0, precalc_fn, potential_fn, fprime, potential, q, synchronicity = args
    max_iter = settings.MAX_SOLVER_ITERS
    precalc_vals = precalc_fn(*(synchronicity, q, phi, z, components_distance), return_as_tuple=True)
    x0 = np.full(phi.shape, x0)
    radius_kwargs = dict(fprime=fprime, maxiter=max_iter, rtol=1e-10, args=((q,) + precalc_vals, potential))
    radius = opt.newton.newton(potential_fn, x0, **radius_kwargs)
    return utils.cylindrical_to_cartesian(np.column_stack((up.abs(radius), phi, z)))


def mesh_detached(system, components_distance, component, symmetry_output=False):
    """
    Creates surface mesh of given binary star component in case of detached or semi-detached system.

    :param system: elisa.binary_system.contaienr.OrbitalPositionContainer;
    :param symmetry_output: bool; if True, besides surface points are returned also `symmetry_vector`,
                                  `base_symmetry_points_number`, `inverse_symmetry_matrix`
    :param component: str; `primary` or `secondary`
    :param components_distance: numpy.float
    :return: Union[Tuple, numpy.array]; (if `symmetry_output` is False)

    Array of surface points if symmetry_output = False::

         numpy.array([[x1 y1 z1],
                      [x2 y2 z2],
                       ...
                      [xN yN zN]])

    othervise::

        (
         numpy.array([[x1 y1 z1],
                      [x2 y2 z2],
                        ...
                      [xN yN zN]]) - array of surface points,
         numpy.float - number of points included in symmetrical one quarter of surface,
         numpy.array([quadrant[indexes_of_remapped_points_in_quadrant]) - matrix of four sub matrices that
                                                                          map basic symmetry quadrant points to all
                                                                          others quadrants
        )
    """
    star = getattr(system, component)
    discretization_factor = star.discretization_factor
    synchronicity = star.synchronicity
    mass_ratio = system.mass_ratio
    potential = star.surface_potential

    potential_fn = getattr(model, f"potential_{component}_fn")
    precalc_fn = getattr(model, f"pre_calculate_for_potential_value_{component}")
    fprime = getattr(model, f"radial_{component}_potential_derivative")

    # pre calculating azimuths for surface points on quarter of the star surface
    phi, theta, separator = pre_calc_azimuths_for_detached_points(discretization_factor, star)
    setattr(star, "azimuth_args", (phi, theta, separator))
    # calculating mesh in cartesian coordinates for quarter of the star, the forward point nearest to the L1 is ommitted
    # it was found that in rare instances, newton performs badly near L1
    args = (phi[1:], theta[1:], star.side_radius, components_distance, precalc_fn) + \
           (potential_fn, fprime, potential, mass_ratio, synchronicity)

    logger.debug(f'calculating surface points of {component} component in mesh_detached '
                 f'function using single process method')
    points_q = get_surface_points(*args)
    # inserting forward surface point from pre-calculated forward radius
    points_q = np.insert(points_q, 0, [star.forward_radius, 0.0, 0.0], axis=0)
    points = stitch_quarters_in_detached(points_q, separator, component, components_distance)

    if symmetry_output:
        equator_length = separator[0] - 2
        meridian_length = separator[1] - separator[0]
        quarter_length = np.shape(points_q)[0] - separator[1]
        base_symmetry_points_number = 2 + equator_length + quarter_length + meridian_length

        points_length = np.shape(points)[0]
        inverse_symmetry_matrix = \
            np.array([up.arange(base_symmetry_points_number),  # 1st quadrant
                      up.concatenate(([0, 1],
                                      up.arange(base_symmetry_points_number + quarter_length,
                                                base_symmetry_points_number + quarter_length + equator_length),
                                      up.arange(base_symmetry_points_number,
                                                base_symmetry_points_number + quarter_length),
                                      up.arange(base_symmetry_points_number - meridian_length,
                                                base_symmetry_points_number))),  # 2nd quadrant
                      up.concatenate(([0, 1],
                                      up.arange(base_symmetry_points_number + quarter_length,
                                                base_symmetry_points_number + quarter_length + equator_length),
                                      up.arange(base_symmetry_points_number + quarter_length + equator_length,
                                                base_symmetry_points_number + 2 * quarter_length + equator_length +
                                                meridian_length))),  # 3rd quadrant
                      up.concatenate((up.arange(2 + equator_length),
                                      up.arange(points_length - quarter_length, points_length),
                                      up.arange(base_symmetry_points_number + 2 * quarter_length + equator_length,
                                                base_symmetry_points_number + 2 * quarter_length + equator_length +
                                                meridian_length)))  # 4th quadrant
                      ])

        return points, base_symmetry_points_number, inverse_symmetry_matrix
    else:
        return points


def rebuild_mesh_detached(system, components_distance, component):
    """
    Rebuilds symmetrical surface mesh of given binary star component in case of detached or semi-detached system.

    :param system: elisa.binary_system.contaienr.OrbitalPositionContainer;
    :param component: str; `primary` or `secondary`
    :param components_distance: numpy.float
    :return: Union[Tuple, numpy.array]; (if `symmetry_output` is False)

    Array of surface points if symmetry_output = False::

         numpy.array([[x1 y1 z1],
                      [x2 y2 z2],
                       ...
                      [xN yN zN]])

    """
    star = getattr(system, component)
    synchronicity = star.synchronicity
    mass_ratio = system.mass_ratio
    potential = star.surface_potential

    if is_empty(star.points):
        raise RuntimeError('This function can be used only on container with already built mesh')
    if star.base_symmetry_points_number == 0:
        raise RuntimeError('This function can be used only on symmetrical meshes')

    potential_fn = getattr(model, f"potential_{component}_fn")
    precalc_fn = getattr(model, f"pre_calculate_for_potential_value_{component}")
    fprime = getattr(model, f"radial_{component}_potential_derivative")

    phi, theta, separator = star.azimuth_args

    args = (phi, theta, star.side_radius, components_distance, precalc_fn) + \
           (potential_fn, fprime, potential, mass_ratio, synchronicity)

    logger.debug(f're calculating surface points of {component} component in rebuild_mesh_detached ')
    points_q = np.round(get_surface_points(*args), 15)
    return stitch_quarters_in_detached(points_q, separator, component, components_distance)


def stitch_quarters_in_detached(points_q, separator, component, components_distance):
    """
    Stitching together surface from symmetrical quarter

    :param points_q: numpy.array; point on the symmetrical quarter including equator and meridian
    :param separator: numpy.array; indices that separates equatorial meridian and inside points
    :param component: str;
    :param components_distance: numpy.float;
    :return: numpy.array; stitched surface
    """
    equator = points_q[:separator[0], :]
    # assigning equator points and nearside and farside points A and B
    x_a, x_eq, x_b = equator[0, 0], equator[1: -1, 0], equator[-1, 0]
    y_a, y_eq, y_b = equator[0, 1], equator[1: -1, 1], equator[-1, 1]
    z_a, z_eq, z_b = equator[0, 2], equator[1: -1, 2], equator[-1, 2]

    # calculating points on phi = 0 meridian
    meridian = points_q[separator[0]: separator[1], :]
    x_meridian, y_meridian, z_meridian = meridian[:, 0], meridian[:, 1], meridian[:, 2]

    # the rest of the surface
    quarter = points_q[separator[1]:, :]
    x_q, y_q, z_q = quarter[:, 0], quarter[:, 1], quarter[:, 2]

    # stiching together 4 quarters of stellar surface in order:
    # north hemisphere: left_quadrant (from companion point of view):
    #                   nearside_point, farside_point, equator, quarter, meridian
    #                   right_quadrant:
    #                   quadrant, equator
    # south hemisphere: right_quadrant:
    #                   quadrant, meridian
    #                   left_quadrant:
    #                   quadrant
    x = np.array([x_a, x_b])
    y = np.array([y_a, y_b])
    z = np.array([z_a, z_b])
    x = up.concatenate((x, x_eq, x_q, x_meridian, x_q, x_eq, x_q, x_meridian, x_q))
    y = up.concatenate((y, y_eq, y_q, y_meridian, -y_q, -y_eq, -y_q, -y_meridian, y_q))
    z = up.concatenate((z, z_eq, z_q, z_meridian, z_q, z_eq, -z_q, -z_meridian, -z_q))

    x = -x + components_distance if component == 'secondary' else x
    return np.column_stack((x, y, z))


def mesh_over_contact(system, component="all", symmetry_output=False):
    """
    Creates surface mesh of given binary star component in case of over-contact system.

    :param system: elisa.binary_system.contaienr.OrbitalPositionContainer;
    :param symmetry_output: bool; if true, besides surface points are returned also `symmetry_vector`,
                                 `base_symmetry_points_number`, `inverse_symmetry_matrix`
    :param component: str; `primary` or `secondary`
    :return: Union[Tuple, numpy.array]; (if symmetry_output is False)

    Array of surface points if symmetry_output = False::

        numpy.array([[x1 y1 z1],
                     [x2 y2 z2],
                      ...
                     [xN yN zN]])

    otherwise::

             numpy.array([[x1 y1 z1],
                          [x2 y2 z2],
                           ...
                          [xN yN zN]]) - array of surface points,

             numpy.float - number of points included in symmetrical one quarter of surface,
             numpy.array([quadrant[indexes_of_remapped_points_in_quadrant]) - matrix of four sub matrices that
                                                                              map base symmetry quadrant to all others
                                                                              quadrants
    """
    star = getattr(system, component)
    discretization_factor = star.discretization_factor
    synchronicity = star.synchronicity
    q = system.mass_ratio
    potential = star.surface_potential
    r_polar = star.polar_radius

    # calculating distance between components
    components_distance = 1.0  # system.orbit.orbital_motion(phase=0)[0][0]

    fn = getattr(model, f"potential_{component}_fn")
    fn_cylindrical = getattr(model, f"potential_{component}_cylindrical_fn")
    precalc = getattr(model, f"pre_calculate_for_potential_value_{component}")
    precal_cylindrical = getattr(model, f"pre_calculate_for_potential_value_{component}_cylindrical")
    fprime = getattr(model, f"radial_{component}_potential_derivative")
    cylindrical_fprime = getattr(model, f"radial_{component}_potential_derivative_cylindrical")

    # precalculating azimuths for farside points and nearside points
    neck_position, neck_polynomial = calculate_neck_position(system, return_polynomial=True)
    (phi_farside, theta_farside, separator_farside), (phi_neck, z_neck, separator_neck) = \
        pre_calc_azimuths_for_overcontact_points(discretization_factor, star, component, neck_position, neck_polynomial)

    # here implement multiprocessing
    args = phi_farside, theta_farside, r_polar, components_distance, precalc, fn, fprime, potential, q, synchronicity
    logger.debug(f'calculating farside points of {component} component in mesh_overcontact '
                 f'function using single process method')
    points_farside = get_surface_points(*args)

    # assigning equator points and point A (the point on the tip of the farside equator)
    equator_farside = points_farside[:separator_farside[0], :]
    x_eq1, x_a = equator_farside[: -1, 0], equator_farside[-1, 0]
    y_eq1, y_a = equator_farside[: -1, 1], equator_farside[-1, 1]
    z_eq1, z_a = equator_farside[: -1, 2], equator_farside[-1, 2]

    # assigning points on phi = pi
    meridian_farside1 = points_farside[separator_farside[0]: separator_farside[1], :]
    x_meridian1, y_meridian1, z_meridian1 = meridian_farside1[:, 0], meridian_farside1[:, 1], meridian_farside1[:, 2]

    # assigning points on phi = pi/2 meridian, perpendicular to component`s distance vector
    meridian_farside2 = points_farside[separator_farside[1]: separator_farside[2], :]
    x_meridian2, y_meridian2, z_meridian2 = meridian_farside2[:, 0], meridian_farside2[:, 1], meridian_farside2[:, 2]

    # assigning the rest of the surface on farside
    quarter = points_farside[separator_farside[2]:, :]
    x_q1, y_q1, z_q1 = quarter[:, 0], quarter[:, 1], quarter[:, 2]

    # solving points on neck
    args = (phi_neck, z_neck, components_distance, 0.25 * star.polar_radius) + \
           (precal_cylindrical, fn_cylindrical, cylindrical_fprime) + \
           (star.surface_potential, system.mass_ratio, synchronicity)
    logger.debug(f'calculating neck points of {component} component in mesh_overcontact '
                 f'function using single process method')
    points_neck = get_surface_points_cylindrical(*args)

    # assigning equator points on neck
    r_eqn = points_neck[:separator_neck[0], :]
    z_eqn, y_eqn, x_eqn = r_eqn[:, 0], r_eqn[:, 1], r_eqn[:, 2]

    # assigning points on phi = 0 meridian, perpendicular to component`s distance vector
    r_meridian_n = points_neck[separator_neck[0]: separator_neck[1], :]
    z_meridian_n, y_meridian_n, x_meridian_n = r_meridian_n[:, 0], r_meridian_n[:, 1], r_meridian_n[:, 2]

    # assigning the rest of the surface on neck
    r_n = points_neck[separator_neck[1]:, :]
    z_n, y_n, x_n = r_n[:, 0], r_n[:, 1], r_n[:, 2]

    # building point blocks similar to those in detached system (equator pts, meridian pts and quarter pts)
    x_eq = up.concatenate((x_eqn, x_eq1), axis=0)
    y_eq = up.concatenate((y_eqn, y_eq1), axis=0)
    z_eq = up.concatenate((z_eqn, z_eq1), axis=0)
    x_q = up.concatenate((x_n, x_meridian2, x_q1), axis=0)
    y_q = up.concatenate((y_n, y_meridian2, y_q1), axis=0)
    z_q = up.concatenate((z_n, z_meridian2, z_q1), axis=0)
    x_meridian = up.concatenate((x_meridian_n, x_meridian1), axis=0)
    y_meridian = up.concatenate((y_meridian_n, y_meridian1), axis=0)
    z_meridian = up.concatenate((z_meridian_n, z_meridian1), axis=0)

    x = up.concatenate((np.array([x_a]), x_eq, x_q, x_meridian, x_q, x_eq, x_q, x_meridian, x_q))
    y = up.concatenate((np.array([y_a]), y_eq, y_q, y_meridian, -y_q, -y_eq, -y_q, -y_meridian, y_q))
    z = up.concatenate((np.array([z_a]), z_eq, z_q, z_meridian, z_q, z_eq, -z_q, -z_meridian, -z_q))

    x = -x + components_distance if component == 'secondary' else x
    points = np.column_stack((x, y, z))
    if symmetry_output:
        equator_length = np.shape(x_eq)[0]
        meridian_length = np.shape(x_meridian)[0]
        quarter_length = np.shape(x_q)[0]
        base_symmetry_points_number = 1 + equator_length + quarter_length + meridian_length

        points_length = np.shape(x)[0]
        inverse_symmetry_matrix = \
            np.array([up.arange(base_symmetry_points_number),  # 1st quadrant
                      up.concatenate(([0],
                                      up.arange(base_symmetry_points_number + quarter_length,
                                                base_symmetry_points_number + quarter_length + equator_length),
                                      up.arange(base_symmetry_points_number,
                                                base_symmetry_points_number + quarter_length),
                                      up.arange(base_symmetry_points_number - meridian_length,
                                                base_symmetry_points_number))),  # 2nd quadrant
                      up.concatenate(([0],
                                      up.arange(base_symmetry_points_number + quarter_length,
                                                base_symmetry_points_number + quarter_length + equator_length),
                                      up.arange(base_symmetry_points_number + quarter_length + equator_length,
                                                base_symmetry_points_number + 2 * quarter_length + equator_length +
                                                meridian_length))),  # 3rd quadrant
                      up.concatenate((up.arange(1 + equator_length),
                                      up.arange(points_length - quarter_length, points_length),
                                      up.arange(base_symmetry_points_number + 2 * quarter_length + equator_length,
                                                base_symmetry_points_number + 2 * quarter_length + equator_length +
                                                meridian_length)))  # 4th quadrant
                      ])

        return points, base_symmetry_points_number, inverse_symmetry_matrix
    else:
        return points


def mesh_spots(system, components_distance, component="all"):
    """
    Compute points of each spots and assigns values to spot container instance.
    If any of any spot point cannot be obtained, entire spot will be omitted.

    :param system: elisa.binary_system.contaienr.OrbitalPositionContainer;
    :param component: str;
    :param components_distance: float;
    :return: bool;
    """

    def solver_condition(x, *_args):
        if isinstance(x, np.ndarray):
            x = x[0]
        point = utils.spherical_to_cartesian([x, _args[1], _args[2]])
        point[0] = point[0] if component == "primary" else components_distance - point[0]
        # ignore also spots where one of points is situated just on the neck
        if getattr(system, "morphology") == "over-contact":
            if (component == "primary" and point[0] >= neck_position) or \
                    (component == "secondary" and point[0] <= neck_position):
                return False
        return True

    components = bsutils.component_to_list(component)
    fns = {
        "primary": (model.potential_primary_fn, model.pre_calculate_for_potential_value_primary,
                    model.radial_primary_potential_derivative),
        "secondary": (model.potential_secondary_fn, model.pre_calculate_for_potential_value_secondary,
                      model.radial_secondary_potential_derivative)
    }
    fns = {component: fns[component] for component in components}

    # in case of wuma system, get separation and make additional test of location of each point (if primary
    # spot doesn't intersect with secondary, if does, then such spot will be skipped completly)
    neck_position = calculate_neck_position(system) if system.morphology == "over-contact" else 1e10

    for component, functions in fns.items():
        logger.debug(f"evaluating spots for {component} component")
        potential_fn, precalc_fn, fprime = functions
        component_instance = getattr(system, component)

        if not component_instance.spots:
            logger.debug(f"no spots to evaluate for {component} component - continue")
            continue

        # iterate over spots
        for spot_index, spot_instance in list(component_instance.spots.items()):
            # lon -> phi, lat -> theta
            lon, lat = spot_instance.longitude, spot_instance.latitude

            alpha = spot_instance.discretization_factor \
                if spot_instance.discretization_factor < spot_instance.angular_radius else spot_instance.angular_radius
            spot_radius = spot_instance.angular_radius
            synchronicity = component_instance.synchronicity
            mass_ratio = system.mass_ratio
            potential = component_instance.surface_potential

            # initial radial vector
            radial_vector = np.array([1.0, lon, lat])  # unit radial vector to the center of current spot
            center_vector = utils.spherical_to_cartesian([1.0, lon, lat])
            args1, use = (synchronicity, mass_ratio, components_distance, radial_vector[1], radial_vector[2]), False
            args2 = ((system.mass_ratio,) + precalc_fn(*args1), potential)
            kwargs = {'original_kwargs': args1}
            solution, use = fsolver(potential_fn, solver_condition, *args2, **kwargs)

            if not use:
                # in case of spots, each point should be usefull, otherwise remove spot from
                # component spot list and skip current spot computation
                if not settings.SUPPRESS_WARNINGS:
                    logger.warning(f"center of spot {spot_instance.kwargs_serializer()} "
                                   f"doesn't satisfy reasonable conditions and entire spot will be omitted")

                component_instance.remove_spot(spot_index=spot_index)
                continue

            spot_center_r = solution
            spot_center = utils.spherical_to_cartesian([spot_center_r, lon, lat])

            # compute euclidean distance of two points on spot (x0)
            # we have to obtain distance between center and 1st point in 1st inner ring of spot
            args1, use = (synchronicity, mass_ratio, components_distance, lon, lat + alpha), False
            args2 = ((system.mass_ratio,) + precalc_fn(*args1), potential)
            kwargs = {'original_kwargs': args1}
            solution, use = fsolver(potential_fn, solver_condition, *args2, **kwargs)

            if not use:
                # in case of spots, each point should be usefull, otherwise remove spot from
                # component spot list and skip current spot computation
                if not settings.SUPPRESS_WARNINGS:
                    logger.warning(f"first inner ring of spot {spot_instance.kwargs_serializer()} "
                                   f"doesn't satisfy reasonable conditions and entire spot will be omitted")

                component_instance.remove_spot(spot_index=spot_index)
                continue

            x0 = up.sqrt(spot_center_r ** 2 + solution ** 2 - (2.0 * spot_center_r * solution * up.cos(alpha)))

            # number of points in latitudal direction
            # + 1 to obtain same discretization as object itself
            num_radial = int(np.round(spot_radius / alpha)) + 1
            logger.debug(f'number of rings in spot {spot_instance.kwargs_serializer()} is {num_radial}')
            thetas = np.linspace(lat, lat + spot_radius, num=num_radial, endpoint=True)

            num_azimuthal = [1 if i == 0 else int(i * 2.0 * const.PI * x0 // x0) for i in range(0, len(thetas))]
            deltas = [np.linspace(0., const.FULL_ARC, num=num, endpoint=False) for num in num_azimuthal]

            spot_phi, spot_theta = [], []
            for theta_index, theta in enumerate(thetas):
                # first point of n-th ring of spot (counting start from center)
                default_spherical_vector = [1.0, lon % const.FULL_ARC, theta]

                for delta_index, delta in enumerate(deltas[theta_index]):
                    # rotating default spherical vector around spot center vector and thus generating concentric
                    # circle of points around centre of spot
                    delta_vector = utils.arbitrary_rotation(theta=delta, omega=center_vector,
                                                            vector=utils.spherical_to_cartesian(
                                                                default_spherical_vector),
                                                            degrees=False,
                                                            omega_normalized=True)

                    spherical_delta_vector = utils.cartesian_to_spherical(delta_vector)

                    spot_phi.append(spherical_delta_vector[1])
                    spot_theta.append(spherical_delta_vector[2])

            spot_phi, spot_theta = np.array(spot_phi), np.array(spot_theta)
            args = (spot_phi, spot_theta, spot_center_r, components_distance, precalc_fn) + \
                   (potential_fn, fprime, potential, mass_ratio, synchronicity)
            try:
                spot_points = get_surface_points(*args)
            except (MaxIterationError, ValueError):
                raise SpotError(f"Solver could not find at least some surface points of spot "
                                f"{spot_instance.kwargs_serializer()}. Probable reason is that your spot is"
                                f"intersecting neck which is currently not supported.")

            if getattr(system, "morphology") == "over-contact":
                if spot_points.ndim == 2:
                    validity_test = (spot_points[:, 0] <= neck_position).all() if component == 'primary' else \
                        (spot_points[:, 0] <= (1 - neck_position)).all()
                else:
                    validity_test = False

                if not validity_test:
                    raise SpotError(f"Your spot {spot_instance.kwargs_serializer()} "
                                    f"is intersecting neck which is currently not supported.")

            boundary_points = spot_points[-len(deltas[-1]):]

            if component == "primary":
                spot_instance.points = np.array(spot_points)
                spot_instance.boundary = np.array(boundary_points)
                spot_instance.center = np.array(spot_center)
            else:
                spot_instance.points = np.array([np.array([components_distance - point[0], -point[1], point[2]])
                                                 for point in spot_points])

                spot_instance.boundary = np.array([np.array([components_distance - point[0], -point[1], point[2]])
                                                   for point in boundary_points])

                spot_instance.center = np.array([components_distance - spot_center[0], -spot_center[1], spot_center[2]])


def calculate_neck_position(system, return_polynomial=False):
    """
    Function calculates x-coordinate of the `neck` (the narrowest place) of an over-contact system.

    :return: Union[Tuple (if return_polynomial is True), float];

    If return_polynomial is set to True::

        (neck position: float, polynomial degree: int)

    otherwise::

        float
    """
    n_points = int(100 * np.radians(5) / system.primary.discretization_factor)
    degree = 15
    components_distance = 1.0

    # solving for neck points
    star = system.primary
    precal_cylindrical = getattr(model, "pre_calculate_for_potential_value_primary_cylindrical")
    fn_cylindrical = getattr(model, "potential_primary_cylindrical_fn")
    cylindrical_fprime = getattr(model, "radial_primary_potential_derivative_cylindrical")

    phi = np.zeros(n_points)
    z = np.linspace(0, 1, num=n_points)
    args = \
        phi, z, components_distance, 0.5 * star.polar_radius, precal_cylindrical, fn_cylindrical, cylindrical_fprime, \
        star.surface_potential, system.mass_ratio, 1.0

    points_neck = get_surface_points_cylindrical(*args)
    x = np.abs(points_neck[:, 2])
    r_c = np.abs(points_neck[:, 0])

    # fiting polynomial to the neck points and searching for neck
    polynomial_fit = np.polyfit(x, r_c, deg=degree)
    polynomial_fit_differentiation = np.polyder(polynomial_fit)
    roots = np.roots(polynomial_fit_differentiation)

    # discarding imaginary solutions
    roots = np.real(roots[np.imag(roots) == 0])

    # choosing root that is closest to the middle of the system, should work...
    # idea is to rule out roots near 0 or 1
    dist_to_cntr = np.abs(roots - 0.5)
    neck_position = roots[np.argmin(dist_to_cntr)]

    if return_polynomial:
        return neck_position, polynomial_fit
    else:
        return neck_position


def add_spots_to_mesh(system, components_distance, component="all"):
    """
    Function implements surface points into clean mesh and removes stellar
    points and other spot points under the given spot if such overlapped spots exists.

    :param system: elisa.binary_system.contaienr.OrbitalPositionContainer;
    :param components_distance: float;
    :param component: Union[str, None];
    """

    components = bsutils.component_to_list(component)

    if is_empty(components):
        # skip building if not required
        return

    component_com = {'primary': 0.0, 'secondary': components_distance}
    for component in components:
        star = getattr(system, component)
        mesh_spots(system, components_distance=components_distance, component=component)
        incorporate_spots_mesh(star, component_com=component_com[component])


def correct_mesh(system, components_distance=None, component='all'):
    """
    Correcting the underestimation of the surface due to the discretization.

    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param components_distance: float; distance of components in SMA units
    :param component: str;
    :return: elisa.binary_system.container.OrbitalPositionContainer;
    """
    components = bsutils.component_to_list(component)

    com = {'primary': 0, 'secondary': components_distance}
    for component in components:
        star = getattr(system, component)
        correct_component_mesh(star, com=com[component], correction_factors=CORRECTION_FACTORS[system.morphology])

    return system



import numpy as np

from typing import Dict, Union
from abc import ABCMeta, abstractmethod
from .. base.body import Body
from .. logger import getLogger
from .. import utils
from .. pulse import pulsations
from elisa import units as u
from elisa.base.surface.temperature import interpolate_bolometric_gravity_darkening
from elisa import const


logger = getLogger('base.system')


class System(metaclass=ABCMeta):
    """
    Abstract class defining System.
    Following arguments are implemented as common any of child instances.
    """

    ID = 1
    MANDATORY_KWARGS = []
    OPTIONAL_KWARGS = []
    ALL_KWARGS = MANDATORY_KWARGS + OPTIONAL_KWARGS

    STAR_MANDATORY_KWARGS = []
    STAR_OPTIONAL_KWARGS = []
    STAR_ALL_KWARGS = STAR_MANDATORY_KWARGS + STAR_OPTIONAL_KWARGS

    def __init__(self, name=None, **kwargs):
        # default params
        self.inclination: float = np.nan
        self.period: float = np.nan
        self.t0: float = np.nan
        self.gamma: float = np.nan
        self.additional_light: float = 0.0
        self.distance: float = 10 * const.PC

        self._components: Union[None, Dict] = None

        if utils.is_empty(name):
            self.name = str(System.ID)
            logger.debug(f"name of class instance {self.__class__.__name__} autoset to {self.name}")
            self.__class__.ID += 1
        else:
            self.name = str(name)

    @property
    @abstractmethod
    def components(self):
        pass

    @abstractmethod
    def compute_lightcurve(self, *args, **kwargs):
        pass

    @abstractmethod
    def init(self):
        pass

    @abstractmethod
    def transform_input(self, *args, **kwargs):
        pass

    @classmethod
    @abstractmethod
    def from_json(cls, data, _verify, _kind_of):
        pass

    @property
    @abstractmethod
    def default_input_units(self):
        pass

    @property
    @abstractmethod
    def default_internal_units(self):
        pass

    def to_json(self):
        """
        Serialize System instance to JSON.

        :return: Dict; JSON serializable
        """

        sys_units: Union[u.DefaultBinarySystemUnits, u.DefaultStarUnits] = self.default_internal_units
        sys_input = self.default_input_units

        spot_units = u.DefaultSpotUnits
        spot_input = u.DefaultSpotInputUnits

        mode_input = u.DefaultPulsationsInputUnits
        mode_units = u.DefaultPulsationsUnits

        json_data = {
            "system": {
                attr: (getattr(self, attr) * sys_units.system[attr]).to(sys_input['system'][attr]).value
                for attr in self.ALL_KWARGS
            }
        }

        for component, instance in self.components.items():
            json_data.update({component: {
                attr: (getattr(instance, attr) * sys_units[component][attr]).to(sys_input[component][attr]).value
                for attr in self.STAR_ALL_KWARGS if getattr(instance, attr) is not None
            }})

            for item in ['atmosphere', 'limb_darkening_coefficients']:
                if getattr(instance, item, None):
                    json_data[component].update({item: getattr(instance, item)})

            if instance.has_spots():
                spot_list = list()
                for spot in instance.spots.values():
                    spot_list.append({
                        attr: (getattr(spot, attr) * spot_units[attr]).to(spot_input[attr]).value
                        for attr in spot.ALL_KWARGS
                    })
                json_data[component].update(dict(spots=spot_list))

            if instance.has_pulsations():
                mode_list = list()
                for mode in instance.pulsations.values():
                    mode_list.append({
                        attr: (getattr(mode, attr) * mode_units[attr]).to(mode_input[attr]).value
                        if attr != 'tidally_locked' else getattr(mode, attr) for attr in mode.ALL_KWARGS
                    })
                json_data[component].update(dict(pulsations=mode_list))

        return json_data

    def assign_pulsations_amplitudes(self, normalisation_constant=1.0):
        """
        Function assigns amplitudes of displacement to each mode based on radial velocity amplitude.

        :param normalisation_constant: float;
        """
        for component, component_instance in self._components.items():
            if component_instance.has_pulsations():
                pulsations.assign_amplitudes(component_instance, normalisation_constant)

    def init_properties(self, **kwargs):
        """
        Setup system properties from input.

        :param kwargs: Dict; all supplied input properties
        """
        logger.debug(f"initialising properties of system {self.name}, values: {kwargs}")
        for kwarg in kwargs:
            setattr(self, kwarg, kwargs[kwarg])

    def has_pulsations(self):
        """
        Resolve whether any of components has pulsation.

        :return: bool;
        """
        return any([instance.has_pulsations() for instance in self.components.values()])

    def has_spots(self):
        """
        Resolve whether any of components has spots.

        :return: bool;
        """
        return any([instance.has_spots() for instance in self.components.values()])

    @staticmethod
    def object_params_validity_check(components, mandatory_kwargs):
        """
        Checking if star instances have all additional attributes set properly.

        :param components: str;
        :param mandatory_kwargs: List;
        """
        for component, component_instance in components.items():
            if not isinstance(component_instance, Body):
                raise TypeError(f"Component `{component}` is not instance of class {Body.__name__}")

        # checking if system components have all mandatory parameters initialised
        missing_kwargs = []
        for component, component_instance in components.items():
            for kwarg in mandatory_kwargs:
                if utils.is_empty(getattr(component_instance, kwarg)):
                    missing_kwargs.append(f"`{kwarg}`")

            if len(missing_kwargs) != 0:
                raise ValueError(f'Mising argument(s): {", ".join(missing_kwargs)} '
                                 f'in {component} component Star class')

    def kwargs_serializer(self):
        """
        Creating dictionary of keyword arguments of System class in order to be able to reinitialize the class
        instance in init().

        :return: Dict;
        """
        cls_name = type(self).__name__
        serialized_kwargs = dict()
        for kwarg in self.ALL_KWARGS:
            value = getattr(self, kwarg)
            if isinstance(value, u.Quantity):
                serialized_kwargs[kwarg] = getattr(self, kwarg)
            else:
                def_unit = getattr(u.default_unit_map[cls_name], kwarg)
                serialized_kwargs[kwarg] = value if def_unit == u.dimensionless_unscaled else value * def_unit
        return serialized_kwargs

    def setup_betas(self):
        """
        Setup of default gravity darkening components.
        """
        for component, instance in self.components.items():
            if utils.is_empty(instance.gravity_darkening):
                instance.gravity_darkening = interpolate_bolometric_gravity_darkening(instance.t_eff)

    setup_gravity_darkening = setup_betas

    @abstractmethod
    def get_positions_method(self, *args, **kwargs):
        pass



import numpy as np
from scipy.interpolate import Akima1DInterpolator

from .. import dynamic
from .. curves import (
    utils as crv_utils,
    c_managed
)
from ... logger import getLogger
from ... import utils, settings, const
from ... binary_system.orbit.container import OrbitalSupplements
from ... observer.mp_manager import manage_observations
from ... binary_system.orbit.orbit import (
    component_distance_from_mean_anomaly,
    get_approx_ecl_angular_width
)


logger = getLogger("binary_system.curves.c_appx_router")


def look_for_approximation(not_pulsations_test):
    """
    This condition checks if even to attempt to utilize apsidal line symmetry approximations.

    :param not_pulsations_test: bool;
    :return: bool;
    """
    valid_nu_separation = settings.MAX_NU_SEPARATION > 0 and settings.MAX_NU_SEPARATION is not None
    valid_max_d_flux = settings.MAX_D_FLUX > 0 and settings.MAX_D_FLUX is not None

    interp_approx = valid_nu_separation and settings.USE_INTERPOLATION_APPROXIMATION
    symmetrical_approx = settings.USE_SYMMETRICAL_COUNTERPARTS_APPROXIMATION
    neighbour_approx = valid_max_d_flux and settings.USE_SIMILAR_NEIGHBOURS_APPROXIMATION
    appx = interp_approx or neighbour_approx or symmetrical_approx
    return appx and not_pulsations_test


def resolve_ecc_approximation_method(
        binary,
        phases,
        position_method,
        try_to_find_appx,
        phases_span_test,
        crv_labels,
        curve_fn,
        **kwargs
):
    """
    Resolve and return approximation method to compute lightcurve in case of eccentric orbit.
    Return value is lambda function with already prepared params.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param phases: numpy.array;
    :param position_method: function;
    :param try_to_find_appx: bool;
    :param phases_span_test: bool; test if phases coverage is sufiicient for phases mirroring along apsidal line
    :param crv_labels: labels of the calculated curves (passbands, components,...)
    :param curve_fn: curve function
    :param kwargs: Dict;
            * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
            * ** left_bandwidth ** * - float
            * ** right_bandwidth ** * - float
    :return: callable;
    """
    approx_method_list = [
        integrate_eccentric_curve_exactly,
        integrate_eccentric_curve_interp_appx,
        integrate_eccentric_curve_symmetrical_counterparts_appx,
        integrate_eccentric_curve_similar_neighbours_appx
    ]

    params = dict(input_argument=phases, return_nparray=True, calculate_from='phase')
    all_orbital_pos_arr = position_method(**params)
    all_orbital_pos = utils.convert_binary_orbital_motion_arr_to_positions(all_orbital_pos_arr)
    potentials = binary.correct_potentials(phases, component="all", iterations=2)

    # APPX ZERO ********************************************************************************************************
    if not try_to_find_appx:
        args = binary, all_orbital_pos, potentials, crv_labels, curve_fn
        return 'zero', lambda: approx_method_list[0](*args, **kwargs)

    # calculating components radii for each orbital position
    radii = crv_utils.forward_radii_from_distances(binary, all_orbital_pos_arr[:, 1], potentials)

    azimuths = all_orbital_pos_arr[:, 2]
    reduced_phase_ids, counterpart_postion_arr, reduced_phase_mask = \
        crv_utils.prepare_apsidaly_symmetric_orbit(binary, azimuths, phases)

    # spliting orbital motion into two separate groups on different sides of apsidal line
    reduced_orbit_arr, reduced_orbit_supplement_arr = \
        crv_utils.split_orbit_by_apse_line(all_orbital_pos_arr, reduced_phase_mask)

    # APPX ONE *********************************************************************************************************
    args = (binary, phases_span_test, reduced_orbit_arr, counterpart_postion_arr, reduced_orbit_supplement_arr)
    interp_appx, reduced_orbit_arr, counterpart_postion_arr = eval_interpolation_approximation(*args)

    if interp_appx:
        args = (binary, radii, phases, reduced_orbit_arr, counterpart_postion_arr, potentials, crv_labels, curve_fn)
        return 'one', lambda: approx_method_list[1](*args, **kwargs)

    # APPX TWO *********************************************************************************************************
    args = binary, radii, reduced_orbit_arr, reduced_orbit_supplement_arr, phases_span_test
    symmetrical_counterparts_appx, orbital_supplements = eval_symmetrical_counterparts_approximation(*args)

    if symmetrical_counterparts_appx:
        args = (binary, radii, phases, orbital_supplements, potentials, crv_labels, curve_fn)
        return 'two', lambda: approx_method_list[2](*args, **kwargs)

    # APPX THREE *******************************************************************************************************
    _args = (binary, radii, all_orbital_pos_arr)
    similar_neighbours_appx, new_geometry_mask, sorted_positions = eval_similar_neighbours_approximation(*_args)
    if similar_neighbours_appx:
        args = (binary, sorted_positions, new_geometry_mask, potentials, crv_labels, curve_fn)
        return 'three', lambda: approx_method_list[3](*args, **kwargs)

    args = binary, all_orbital_pos, potentials, crv_labels, curve_fn
    return 'zero', lambda: approx_method_list[0](*args, **kwargs)


# *******************************************evaluate_approximations****************************************************
# noinspection PyUnresolvedReferences
def eval_interpolation_approximation(
        binary,
        phases_span_test,
        reduced_orbit_array,
        counterpart_position_array,
        reduced_orbit_supplement_arr
):
    """
    Test if it is possible to compute eccentric binary system with interpolation approximation.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param phases_span_test: bool; test for sufficient phase span of observations
    :param reduced_orbit_array: numpy.array; orbital positions defined by user on one side of the apsidal line
    :param counterpart_position_array: numpy.array; symmetrical counterparts to the 'reduced_orbit_array'
    :param reduced_orbit_supplement_arr: numpy.array; orbital positions defined by user on the opposite side as
                                         'reduced_orbit_array' side of the apsidal line
    :return: bool;
    """
    # true anomalies of orbital positions modelled by interpolation approximation
    true_anomalies_supplements = reduced_orbit_supplement_arr[:, 3]

    # base test to establish, if curve contains enough points
    max_nu_sep = np.max(np.diff(np.sort(true_anomalies_supplements)))
    if max_nu_sep > settings.MAX_NU_SEPARATION or 0 > settings.MAX_NU_SEPARATION or not phases_span_test or \
            not settings.USE_INTERPOLATION_APPROXIMATION:
        logger.debug('Orbit is not sufficiently populated to implement interpolation approximation 1')
        return False, reduced_orbit_array, counterpart_position_array

    # component distance during eclipses
    ecl_true_anomalies = np.array([binary.orbit.conjunctions[f'{component}_eclipse']['true_anomaly']
                                  for component in settings.BINARY_COUNTERPARTS])
    distances_at_ecl = component_distance_from_mean_anomaly(binary.eccentricity, ecl_true_anomalies)

    # angular width of each eclipse
    angular_ecl_widths = [get_approx_ecl_angular_width(binary.primary.forward_radius,
                                                       binary.secondary.forward_radius,
                                                       distance, binary.inclination)
                          for distance in distances_at_ecl]

    for ii, ecl_nu in enumerate(ecl_true_anomalies):
        if angular_ecl_widths[ii][0] == 0.0:
            continue

        # including adjacent points to the eclipse to ensure smoothness
        d_nu1 = crv_utils.adjust_eclipse_width(true_anomalies_supplements, ecl_nu - angular_ecl_widths[ii][0])
        d_nu2 = crv_utils.adjust_eclipse_width(true_anomalies_supplements, ecl_nu + angular_ecl_widths[ii][0])

        bottom, top = ecl_nu - angular_ecl_widths[ii][0] - d_nu1, ecl_nu + angular_ecl_widths[ii][0] + d_nu2
        points_ecl_mask_suplements = np.logical_and(true_anomalies_supplements > bottom,
                                                    true_anomalies_supplements < top)
        # treating eclipses on boundaries of 0, 2pi interval
        if bottom < 0.0:
            points_ecl_mask_suplements = np.logical_or(points_ecl_mask_suplements,
                                                       true_anomalies_supplements > bottom + const.FULL_ARC)
        elif top > const.FULL_ARC:
            points_ecl_mask_suplements = np.logical_or(points_ecl_mask_suplements,
                                                       true_anomalies_supplements < top - const.FULL_ARC)

        # number of points in eclipse
        points_in_ecl_suplements = np.sum(points_ecl_mask_suplements)

        # subtraction of the central plateau (taking into account only descent and ascent part)
        plateau_factor = 1 - angular_ecl_widths[ii][1] / angular_ecl_widths[ii][0]

        if plateau_factor * points_in_ecl_suplements < settings.MIN_POINTS_IN_ECLIPSE:
            rowstack_args = (reduced_orbit_array, reduced_orbit_supplement_arr[points_ecl_mask_suplements])
            reduced_orbit_array = np.row_stack(rowstack_args)

            rowstack_args = (counterpart_position_array, np.full((points_in_ecl_suplements, 5), np.nan))
            counterpart_position_array = np.row_stack(rowstack_args)
        else:
            # interpolation approximation causes artifacts in case of the very
            # flat plateaus in the bottom of the eclipse
            return False, reduced_orbit_array, counterpart_position_array

    # removing duplicite entries
    _, indices = np.unique(reduced_orbit_array[:, 0], return_index=True)
    reduced_orbit_array = reduced_orbit_array[indices]
    counterpart_position_array = counterpart_position_array[indices]

    return True, reduced_orbit_array, counterpart_position_array


def eval_symmetrical_counterparts_approximation(binary, radii, base_orbit_arr, orbit_supplement_arr, phases_span_test):
    """
    Test if it is possible to compute eccentric binary system with symmetrical counterparts approximation.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param radii: numpy.array; forward potentials
    :param base_orbit_arr: numpy.array;
    :param orbit_supplement_arr: numpy.array;
    :param phases_span_test: bool;
    :return: Tuple; approximation test, orbital supplements
    """
    if not phases_span_test or not settings.USE_SYMMETRICAL_COUNTERPARTS_APPROXIMATION:
        logger.debug('Phase span of the observation is not sufficient to '
                     'utilize symmetrical counterparts approximation.')
        return False, None

    # create object of separated objects and supplements to bodies
    args = (binary, radii, base_orbit_arr, orbit_supplement_arr)
    orbital_supplements = dynamic.find_apsidally_corresponding_positions(*args)

    orbital_supplements.sort(by='distance')

    # checking how much mirrors are used
    idxs_where_nan = np.argwhere(np.isnan(orbital_supplements.mirror[:, 0]))
    if idxs_where_nan.shape[0] == orbital_supplements.mirror.shape[0]:
        return False, None

    return True, orbital_supplements


def eval_similar_neighbours_approximation(binary, radii, all_orbital_pos_arr):
    """
    Test if it is possible to compute eccentric binary system with similar neighbours approximation.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param radii: numpy.array; forward radii for all_orbital_pos_arr)
    :param all_orbital_pos_arr: numpy.array; array of all orbital positions
    :return: Tuple; (bool, numpy.array), approximation test, new_geometry_test
    """
    if not settings.USE_SIMILAR_NEIGHBOURS_APPROXIMATION:
        return False, None, None

    sort_idxs = all_orbital_pos_arr[:, 1].argsort()
    sorted_all_orbital_pos_arr = all_orbital_pos_arr[sort_idxs]
    sorted_radii = radii[:, sort_idxs]

    rel_d_radii = crv_utils.compute_rel_d_geometry(binary, sorted_radii[:, 1:], sorted_radii[:, 1:])
    args = (binary.has_spots(), all_orbital_pos_arr.shape[0], rel_d_radii)
    new_geometry_mask = dynamic.resolve_object_geometry_update(*args)

    rel_irrad = crv_utils.compute_rel_d_irradiation(binary, all_orbital_pos_arr[:, 1])
    new_irrad_mask = dynamic.resolve_irrad_update(rel_irrad, all_orbital_pos_arr.shape[0])
    new_build_mask = np.logical_or(new_geometry_mask, new_irrad_mask)

    approx_test = not new_build_mask.all()
    return approx_test, new_build_mask, sorted_all_orbital_pos_arr


# *******************************************approximation curve_methods************************************************
def integrate_eccentric_curve_interp_appx(
        binary,
        radii,
        phases,
        reduced_orbit_arr,
        counterpart_postion_arr,
        potentials,
        crv_labels,
        curve_fn,
        **kwargs
):
    """
    Function calculates curves for eccentric orbits for selected filters using approximation
    where curve points on the one side of the apsidal line are calculated exactly and the second
    half of the curve points are calculated by mirroring the surface geometries of the first
    half of the points to the other side of the apsidal line. Since those mirrored points are not
    usually alligned with desired phases, the fluxes for each user defined phase is interpolated.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param radii: numpy.array; forward radii
    :param phases: numpy.array;
    :param reduced_orbit_arr: numpy.array; base orbital positions
    :param counterpart_postion_arr: numpy.array; orbital positions symmetric to the `base_orbit_arr`
    :param potentials: Dict; corrected potentials
    :param crv_labels: List; curve_labels
    :param curve_fn: curve integrator function
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** - float
        * ** right_bandwidth ** - float
        * ** atlas ** - str
    :return: Dict[str, numpy.array];
    """
    n = 5 if phases.shape[0] > 10 else int(phases.shape[0] / 2) - 1
    orbital_supplements = OrbitalSupplements(body=reduced_orbit_arr, mirror=counterpart_postion_arr)
    orbital_supplements = orbital_supplements.sort(by='distance')

    orbital_positions = np.stack((orbital_supplements.body, orbital_supplements.mirror), axis=1)
    fn_args = (binary, potentials, radii, crv_labels, curve_fn)
    fn = c_managed.integrate_eccentric_curve_w_orbital_symmetry
    stacked_band_curves = manage_observations(fn=fn, fn_args=fn_args, position=orbital_positions, **kwargs)

    # interpolation of the points in the second half of the light curves using splines
    x = np.concatenate((orbital_supplements.body[:, 4], orbital_supplements.mirror[:, 4]))
    not_nan_test = ~np.isnan(x)
    x = x[not_nan_test] % 1

    # checking for accidental allignement between templates and mirrors (corresponding to eclipse regions transferred
    # from mirrored orbit area)
    x, uniqe_idx = np.unique(x, return_index=True)

    sort_idx = np.argsort(x)
    x = x[sort_idx]
    x = np.concatenate((x[-n:] - 1, x, x[:n] + 1))

    band_curves = dict()
    for curve in crv_labels:
        y = np.concatenate((stacked_band_curves[curve][:, 0], stacked_band_curves[curve][:, 1]))
        y = ((y[not_nan_test])[uniqe_idx])[sort_idx]
        y = np.concatenate((y[-n:], y, y[:n]))

        i = Akima1DInterpolator(x, y)
        f = i(phases)
        band_curves[curve] = f

    return band_curves


def integrate_eccentric_curve_symmetrical_counterparts_appx(
        binary,
        radii,
        phases,
        orbital_supplements,
        potentials,
        crv_labels,
        curve_fn,
        **kwargs
):
    """
    Function calculates curve for eccentric orbit using
    approximation where to each OrbitalPosition on one side of the apsidal line,
    the closest counterpart OrbitalPosition is assigned and the same surface geometry is
    assumed for both of them.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param radii: numpy.array; forward radii
    :param phases: numpy.array;
    :param orbital_supplements: elisa.binary_system.orbit.container.OrbitalSupplements;
    :param potentials: Dict; corrected potentials
    :param crv_labels: List; curve_labels
    :param curve_fn: curve integrator function
    :param kwargs: Dict;
            * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
            * ** left_bandwidth ** * - float
            * ** right_bandwidth ** * - float
    :return: Dict[str, numpy.array];
    """
    orbital_positions = np.stack((orbital_supplements.body, orbital_supplements.mirror), axis=1)
    fn_args = (binary, potentials, radii, crv_labels, curve_fn)
    fn = c_managed.integrate_eccentric_curve_w_orbital_symmetry
    stacked_band_curves = manage_observations(fn=fn, fn_args=fn_args, position=orbital_positions, **kwargs)

    # re-arranging points to original order
    band_curves = {key: np.empty(phases.shape) for key in crv_labels}
    for lbl in crv_labels:
        base_idxs = np.array(orbital_supplements.body[:, 0], dtype=np.int32)
        band_curves[lbl][base_idxs] = stacked_band_curves[lbl][:, 0]
        not_nan_test = ~np.isnan(orbital_supplements.mirror[:, 0])
        mirror_idxs = np.array(orbital_supplements.mirror[not_nan_test, 0], dtype=np.int32)
        band_curves[lbl][mirror_idxs] = stacked_band_curves[lbl][not_nan_test, 1]

    return band_curves


def integrate_eccentric_curve_similar_neighbours_appx(
        binary,
        orbital_positions,
        new_geometry_mask,
        potentials,
        crv_labels,
        curve_fn,
        **kwargs
):
    """
    Function calculates curve for eccentric orbit using approximation where surface
    is not fully recalculated between sufficiently similar neighbouring orbital positions.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param orbital_positions: numpy.array; orbital positions sorted by components distance
    :param new_geometry_mask: bool; mask to `orbital_positions` which determines which surface
                                    geometry should be fully recalculated
    :param potentials: Dict; corrected surface potentials
    :param crv_labels: List; curve_labels
    :param curve_fn: curve integrator function
    :param kwargs: kwargs: Dict;
            * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
            * ** left_bandwidth ** * - float
            * ** right_bandwidth ** * - float
    :return: Dict[str, numpy.array];
    """
    fn_args = (binary, potentials, new_geometry_mask, crv_labels, curve_fn)
    fn = c_managed.similar_neighbour_approximation_ecc_curve_integration
    band_curves_unsorted = manage_observations(fn=fn, fn_args=fn_args, position=orbital_positions, **kwargs)

    # re-arranging points to original order
    return {key: band_curves_unsorted[key][orbital_positions[:, 0].argsort()] for key in crv_labels}


def integrate_eccentric_curve_exactly(binary, orbital_motion, potentials, crv_labels, curve_fn, **kwargs):
    """
    Function calculates curves for eccentric orbit for selected filters.
    Curve is calculated exactly for each OrbitalPosition.
    It is very slow and it should be used only as a benchmark.

    :param binary: elisa.binary_system.system.BinarySystem; instance
    :param orbital_motion: list of all OrbitalPositions at which curve will be calculated
    :param potentials: Dict; corrected potentials
    :param kwargs: kwargs taken from `produce_eccentric_curve` function
    :param crv_labels: labels of the calculated curves (passbands, components,...)
    :param curve_fn: curve function
    :return: Dict; dictionary of fluxes for each filter
    """
    # surface potentials with constant volume of components
    fn_args = (binary, potentials, None, crv_labels, curve_fn)
    fn = c_managed.integrate_eccentric_curve_exactly
    curves = manage_observations(fn=fn, fn_args=fn_args, position=orbital_motion, **kwargs)
    return curves
